import dataclasses
import datetime
import os
from collections import namedtuple
from collections.abc import Iterable, Mapping
from typing import Literal
from unittest.mock import MagicMock, patch

import cftime
import numpy as np
import pytest
import torch
import xarray as xr

import fme
from fme.ace.aggregator import OneStepAggregator
from fme.ace.aggregator.plotting import plot_paneled_data
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.registry.registry import ModuleConfig
from fme.ace.registry.sfno import SphericalFourierNeuralOperatorBuilder
from fme.ace.stepper.single_module import (
    AtmosphereCorrectorConfig,
    SingleModuleStepperConfig,
    Stepper,
    TrainOutput,
)
from fme.core import AtmosphereData
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.generics.optimization import OptimizationABC
from fme.core.loss import WeightedMappingLossConfig
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.ocean import OceanConfig, SlabOceanConfig
from fme.core.optimization import (
    CheckpointConfig,
    NullOptimization,
    Optimization,
    OptimizationConfig,
)
from fme.core.registry.module import ModuleSelector
from fme.core.testing.regression import validate_tensor_dict

from .test_single_module import get_data, get_dataset_info, get_scalar_data

DIR = os.path.abspath(os.path.dirname(__file__))

SphericalData = namedtuple("SphericalData", ["data", "area_weights", "vertical_coord"])
TIMESTEP = datetime.timedelta(hours=6)
DEVICE = fme.get_device()
OCEAN_CONFIG = OceanConfig(surface_temperature_name="a", ocean_fraction_name="b")
MULTI_CALL_CONFIG = MultiCallConfig(
    forcing_name="co2", forcing_multipliers={"_b": 0.5}, output_names=["ULWRFtoa"]
)
LOAD_STEPPER_TESTS = {
    "override-ocean": (None, None, OCEAN_CONFIG, "keep", OCEAN_CONFIG, None),
    "persist-ocean": (OCEAN_CONFIG, None, "keep", "keep", OCEAN_CONFIG, None),
    "override-multi-call": (
        None,
        None,
        "keep",
        MULTI_CALL_CONFIG,
        None,
        MULTI_CALL_CONFIG,
    ),
    "persist-multi-call": (
        None,
        MULTI_CALL_CONFIG,
        "keep",
        "keep",
        None,
        MULTI_CALL_CONFIG,
    ),
    "override-all": (
        None,
        None,
        OCEAN_CONFIG,
        MULTI_CALL_CONFIG,
        OCEAN_CONFIG,
        MULTI_CALL_CONFIG,
    ),
    "persist-all": (
        OCEAN_CONFIG,
        MULTI_CALL_CONFIG,
        "keep",
        "keep",
        OCEAN_CONFIG,
        MULTI_CALL_CONFIG,
    ),
}


@pytest.mark.parametrize(
    "in_names,out_names,ocean_config,expected_all_names",
    [
        (["a"], ["b"], None, ["a", "b"]),
        (["a"], ["a", "b"], None, ["a", "b"]),
        (["a", "b"], ["b"], None, ["a", "b"]),
        (["a", "b"], ["a", "b"], None, ["a", "b"]),
        (
            ["a", "b"],
            ["a", "b"],
            OceanConfig("a", "mask"),
            ["a", "b", "mask"],
        ),
        (
            ["a", "b"],
            ["a", "b"],
            OceanConfig("a", "b"),
            ["a", "b"],
        ),
        (
            ["a", "b"],
            ["a", "b"],
            OceanConfig("a", "of", False, SlabOceanConfig("c", "d")),
            ["a", "b", "of", "c", "d"],
        ),
    ],
)
def test_stepper_config_all_names_property(
    in_names, out_names, ocean_config, expected_all_names
):
    config = SingleModuleStepperConfig(
        builder=MagicMock(),
        in_names=in_names,
        out_names=out_names,
        normalization=MagicMock(),
        ocean=ocean_config,
    )
    # check there are no duplications
    assert len(config.all_names) == len(set(config.all_names))
    # check the right items are in there using sets to ignore order
    assert set(config.all_names) == set(expected_all_names)


def test_train_on_batch_normalizer_changes_only_norm_data():
    torch.manual_seed(0)
    data = get_data(["a", "b"], n_samples=5, n_time=2).data
    normalization_config = NormalizationConfig(
        means=get_scalar_data(["a", "b"], 0.0),
        stds=get_scalar_data(["a", "b"], 1.0),
    )
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": torch.nn.Identity()}),
        in_names=["a", "b"],
        out_names=["a", "b"],
        normalization=normalization_config,
        loss=WeightedMappingLossConfig(type="MSE"),
    )
    dataset_info = get_dataset_info()
    stepper = config.get_stepper(dataset_info=dataset_info)
    stepped = stepper.train_on_batch(data=data, optimization=NullOptimization())
    assert torch.allclose(
        stepped.gen_data["a"], stepped.normalize(stepped.gen_data)["a"]
    )  # as std=1, mean=0, no change
    normalization_config.stds = get_scalar_data(["a", "b"], 2.0)
    config.normalization = normalization_config
    config.loss_normalization = NormalizationConfig(
        means=get_scalar_data(["a", "b"], 0.0),
        stds=get_scalar_data(["a", "b"], 3.0),
    )
    stepper = config.get_stepper(
        dataset_info=dataset_info,
    )
    stepped_double_std = stepper.train_on_batch(
        data=data, optimization=NullOptimization()
    )
    assert torch.allclose(
        stepped.gen_data["a"], stepped_double_std.gen_data["a"], rtol=1e-4
    )
    assert torch.allclose(
        stepped.gen_data["a"],
        2.0 * stepped_double_std.normalize(stepped_double_std.gen_data)["a"],
        rtol=1e-4,
    )
    assert torch.allclose(
        stepped.target_data["a"],
        2.0 * stepped_double_std.normalize(stepped_double_std.target_data)["a"],
        rtol=1e-4,
    )
    assert torch.allclose(
        stepped.metrics["loss"], 9.0 * stepped_double_std.metrics["loss"], rtol=1e-4
    )  # mse scales with std**2


def test_train_on_batch_addition_series():
    torch.manual_seed(0)

    class AddOne(torch.nn.Module):
        def forward(self, x):
            return x + 1

    n_steps = 4
    data_with_ic: BatchData = get_data(["a", "b"], n_samples=5, n_time=n_steps + 1).data
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": AddOne()}),
        in_names=["a", "b"],
        out_names=["a", "b"],
        normalization=NormalizationConfig(
            means=get_scalar_data(["a", "b"], 0.0),
            stds=get_scalar_data(["a", "b"], 1.0),
        ),
        loss=WeightedMappingLossConfig(type="MSE"),
    )
    dataset_info = get_dataset_info()
    stepper = config.get_stepper(dataset_info=dataset_info)
    stepped = stepper.train_on_batch(data=data_with_ic, optimization=NullOptimization())
    # output of train_on_batch does not include the initial condition
    assert stepped.gen_data["a"].shape == (5, 1, n_steps + 1, 5, 5)

    for i in range(n_steps - 1):
        assert torch.allclose(
            stepped.normalize(stepped.gen_data)["a"][:, :, i] + 1,
            stepped.normalize(stepped.gen_data)["a"][:, :, i + 1],
        )
        assert torch.allclose(
            stepped.normalize(stepped.gen_data)["b"][:, :, i] + 1,
            stepped.normalize(stepped.gen_data)["b"][:, :, i + 1],
        )
        assert torch.allclose(
            stepped.gen_data["a"][:, :, i] + 1, stepped.gen_data["a"][:, :, i + 1]
        )
        assert torch.allclose(
            stepped.gen_data["b"][:, :, i] + 1, stepped.gen_data["b"][:, :, i + 1]
        )
    assert torch.allclose(
        stepped.normalize(stepped.target_data)["a"],
        data_with_ic.data["a"][:, None],
    )
    assert torch.allclose(
        stepped.normalize(stepped.target_data)["b"],
        data_with_ic.data["b"][:, None],
    )


def test_train_on_batch_crps_loss():
    torch.manual_seed(0)

    class AddOne(torch.nn.Module):
        def forward(self, x):
            return x + 1

    n_steps = 4
    data_with_ic: BatchData = get_data(["a", "b"], n_samples=5, n_time=n_steps + 1).data
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": AddOne()}),
        in_names=["a", "b"],
        out_names=["a", "b"],
        normalization=NormalizationConfig(
            means=get_scalar_data(["a", "b"], 0.0),
            stds=get_scalar_data(["a", "b"], 1.0),
        ),
        loss=WeightedMappingLossConfig(type="MSE"),
        crps_training=True,
    )
    dataset_info = get_dataset_info()
    stepper = config.get_stepper(dataset_info=dataset_info)
    stepped = stepper.train_on_batch(data=data_with_ic, optimization=NullOptimization())
    # output of train_on_batch does not include the initial condition
    assert stepped.gen_data["a"].shape == (5, 2, n_steps + 1, 5, 5)


def test_train_on_batch_with_prescribed_ocean():
    torch.manual_seed(0)

    class AddOne(torch.nn.Module):
        def forward(self, x):
            return x + 1

    n_steps = 3
    data: BatchData = get_data(["a", "b", "mask"], n_samples=5, n_time=n_steps + 1).data
    data.data["mask"][:] = 0
    data.data["mask"][:, :, :, 0] = 1
    stds = {"a": 2.0, "b": 3.0}
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": AddOne()}),
        in_names=["a", "b"],
        out_names=["a", "b"],
        normalization=NormalizationConfig(
            means=get_scalar_data(["a", "b"], 0.0),
            stds=stds,
        ),
        ocean=OceanConfig("b", "mask"),
    )
    dataset_info = get_dataset_info()
    stepper = config.get_stepper(dataset_info=dataset_info)
    stepped = stepper.train_on_batch(data, optimization=NullOptimization())
    for i in range(n_steps - 1):
        # "a" should be increasing by 1 according to AddOne
        torch.testing.assert_close(
            stepped.normalize(stepped.gen_data)["a"][:, :, i] + 1,
            stepped.normalize(stepped.gen_data)["a"][:, :, i + 1],
        )
        # "b" should be increasing by 1 where the mask says don't prescribe
        # note the 1: selection for the last dimension in following two assertions
        torch.testing.assert_close(
            stepped.normalize(stepped.gen_data)["b"][:, :, i, :, 1:] + 1,
            stepped.normalize(stepped.gen_data)["b"][:, :, i + 1, :, 1:],
        )
        # now check that the 0th index in last dimension has been overwritten
        torch.testing.assert_close(
            stepped.normalize(stepped.gen_data)["b"][:, :, i, :, 0],
            stepped.normalize({"b": stepped.target_data["b"]})["b"][:, :, i, :, 0],
        )


def test_reloaded_stepper_gives_same_prediction():
    torch.manual_seed(0)
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(
            type="SphericalFourierNeuralOperatorNet", config={"scale_factor": 1}
        ),
        in_names=["a", "b"],
        out_names=["a", "b"],
        normalization=NormalizationConfig(
            means={"a": 0.0, "b": 0.0},
            stds={"a": 1.0, "b": 1.0},
        ),
    )
    dataset_info = get_dataset_info()
    stepper = config.get_stepper(dataset_info=dataset_info)
    new_stepper = Stepper.from_state(stepper.get_state())
    data = get_data(["a", "b"], n_samples=5, n_time=2).data
    first_result = stepper.train_on_batch(
        data=data,
        optimization=NullOptimization(),
    )
    second_result = new_stepper.train_on_batch(
        data=data,
        optimization=NullOptimization(),
    )
    assert torch.allclose(first_result.metrics["loss"], second_result.metrics["loss"])
    assert torch.allclose(first_result.gen_data["a"], second_result.gen_data["a"])
    assert torch.allclose(first_result.gen_data["b"], second_result.gen_data["b"])
    assert torch.allclose(first_result.target_data["a"], second_result.target_data["a"])
    assert torch.allclose(first_result.target_data["b"], second_result.target_data["b"])


class ReturnZerosModule(torch.nn.Module):
    """
    Returns zeros with the correct number of out channels. Creates an unused
    parameter so that optimization has something to gnaw on.
    """

    def __init__(self, n_in_channels, n_out_channels) -> None:
        super().__init__()
        self.n_in_channels = n_in_channels
        self.n_out_channels = n_out_channels
        self._param = torch.nn.Parameter(
            torch.tensor(0.0, device=get_device())
        )  # unused

    def forward(self, x):
        assert torch.all(~torch.isnan(x))
        batch_size, n_channels, nlat, nlon = x.shape
        assert n_channels == self.n_in_channels
        zero = torch.zeros(
            batch_size, self.n_out_channels, nlat, nlon, device=get_device()
        )
        return zero + self._param


def _setup_and_train_on_batch(
    data: BatchData,
    in_names,
    out_names,
    ocean_config: OceanConfig | None,
    optimization_config: OptimizationConfig | None,
    stepper_config_kwargs,
):
    """Sets up the requisite classes to run train_on_batch."""
    module = ReturnZerosModule(len(in_names), len(out_names))

    if optimization_config is None:
        optimization: NullOptimization | Optimization = NullOptimization()
    else:
        optimization = optimization_config.build(modules=[module], max_epochs=2)

    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": module}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means=get_scalar_data(set(in_names + out_names), 0.0),
            stds=get_scalar_data(set(in_names + out_names), 1.0),
        ),
        ocean=ocean_config,
        **stepper_config_kwargs,
    )
    stepper = config.get_stepper(dataset_info=get_dataset_info())
    return stepper.train_on_batch(data, optimization=optimization)


@pytest.mark.parametrize(
    "is_input,is_output,is_prescribed",
    [
        pytest.param(True, True, True, id="in_out_prescribed"),
        pytest.param(True, True, False, id="in_out_not_prescribed"),
        pytest.param(False, True, False, id="out_only_not_prescribed"),
    ],
)
@pytest.mark.parametrize("n_forward_steps", [1, 2, 3], ids=lambda p: f"k={p}")
@pytest.mark.parametrize("is_train", [True, False], ids=["is_train", ""])
@pytest.mark.parametrize(
    "with_activation_checkpointing", [True, False], ids=["act_ckpt", ""]
)
def test_train_on_batch(
    n_forward_steps,
    is_input,
    is_output,
    is_train,
    is_prescribed,
    with_activation_checkpointing,
):
    in_names, out_names = ["a"], ["a"]
    if is_input:
        in_names.append("b")
    if is_output:
        out_names.append("b")
    all_names = sorted(list(set(in_names).union(set(out_names))))

    if is_prescribed:
        mask_name = "mask"
        all_names.append(mask_name)
        in_names.append(mask_name)
        ocean_config = OceanConfig("b", mask_name)
    else:
        ocean_config = None

    data, _, _ = get_data(all_names, 3, n_forward_steps + 1)

    if is_train:
        if with_activation_checkpointing:
            optimization = OptimizationConfig(
                checkpoint=CheckpointConfig(after_n_forward_steps=n_forward_steps - 1)
            )
        else:
            optimization = OptimizationConfig()
    else:
        optimization = None

    with patch("torch.utils.checkpoint.checkpoint") as mock_checkpoint:
        # have the mock call the module and return the step
        mock_checkpoint.side_effect = lambda f, x, **_: f(x)
        _setup_and_train_on_batch(
            data, in_names, out_names, ocean_config, optimization, {}
        )

        if is_train and with_activation_checkpointing:
            # should be called exactly once, for the final forward step
            mock_checkpoint.assert_called_once()
        else:
            mock_checkpoint.assert_not_called()


@pytest.mark.parametrize("n_forward_steps", [1, 2, 3])
def test_train_on_batch_one_step_aggregator(n_forward_steps):
    in_names, out_names, all_names = ["a"], ["a"], ["a"]
    data, _, _ = get_data(all_names, 3, n_forward_steps + 1)
    nx, ny = 5, 5
    stepper = _get_stepper(in_names, out_names, ocean_config=None, module_name="AddOne")
    lat_lon_coordinates = LatLonCoordinates(torch.arange(nx), torch.arange(ny))
    # keep area weights ones for simplicity
    lat_lon_coordinates._area_weights = torch.ones(nx, ny)
    ds_info = DatasetInfo(horizontal_coordinates=lat_lon_coordinates)
    aggregator = OneStepAggregator(ds_info, save_diagnostics=False)

    stepped = stepper.train_on_batch(data, optimization=NullOptimization())
    assert stepped.gen_data["a"].shape[2] == n_forward_steps + 1

    aggregator.record_batch(stepped)
    logs = aggregator.get_logs("one_step")

    gen = data.data["a"].select(dim=1, index=0) + 1
    tar = data.data["a"].select(dim=1, index=1)

    bias = torch.mean(gen - tar)
    assert np.isclose(bias.item(), logs["one_step/mean/weighted_bias/a"])

    residual_gen = torch.ones((5, 5))
    residual_tar = tar[0] - data.data["a"].select(dim=1, index=0)[0]
    residual_imgs = [[residual_gen.cpu().numpy()], [residual_tar.cpu().numpy()]]
    residual_plot = plot_paneled_data(residual_imgs, diverging=True)
    assert np.allclose(
        residual_plot.to_data_array(),
        logs["one_step/snapshot/image-residual/a"].to_data_array(),
    )

    full_field_gen = gen.mean(dim=0)
    full_field_tar = tar.mean(dim=0)
    full_field_plot = plot_paneled_data(
        [
            [full_field_gen.cpu().numpy()],
            [full_field_tar.cpu().numpy()],
        ],
        diverging=False,
    )
    assert np.allclose(
        full_field_plot.to_data_array(),
        logs["one_step/mean_map/image-full-field/a"].to_data_array(),
    )


class Multiply(torch.nn.Module):
    def __init__(self, factor):
        super().__init__()
        self.factor = factor

    def forward(self, x):
        return x * self.factor


@pytest.mark.parametrize(
    "global_only, terms_to_modify, force_positive",
    [
        (True, None, False),
        (True, "precipitation", False),
        (True, "evaporation", False),
        (False, "advection_and_precipitation", False),
        (False, "advection_and_evaporation", False),
        (False, "advection_and_precipitation", True),
    ],
)
@pytest.mark.parametrize("compute_derived_in_train_on_batch", [False, True])
def test_stepper_corrector(
    global_only: bool,
    terms_to_modify,
    force_positive: bool,
    compute_derived_in_train_on_batch: bool,
):
    torch.random.manual_seed(0)
    n_forward_steps = 5
    device = get_device()
    data = {
        "PRESsfc": 10.0 + torch.rand(size=(3, n_forward_steps + 1, 5, 5)),
        "specific_total_water_0": -0.2
        + torch.rand(size=(3, n_forward_steps + 1, 5, 5)),
        "specific_total_water_1": torch.rand(size=(3, n_forward_steps + 1, 5, 5)),
        "PRATEsfc": torch.rand(size=(3, n_forward_steps + 1, 5, 5)),
        "LHTFLsfc": torch.rand(size=(3, n_forward_steps + 1, 5, 5)),
        "tendency_of_total_water_path_due_to_advection": torch.rand(
            size=(3, n_forward_steps + 1, 5, 5)
        ),
    }
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.asarray([3.0, 1.0, 0.0]), bk=torch.asarray([0.0, 0.6, 1.0])
    ).to(device)
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.linspace(-89.5, 89.5, 5, device=device),
        lon=torch.linspace(-179.5, 179.5, 5, device=device),
    )
    dataset_info = get_dataset_info(
        vertical_coordinate=vertical_coordinate,
        horizontal_coordinate=horizontal_coordinate,
    )
    gridded_ops = dataset_info.gridded_operations

    if force_positive:
        force_positive_names = ["specific_total_water_0"]
    else:
        force_positive_names = []

    corrector_config = AtmosphereCorrectorConfig(
        conserve_dry_air=True,
        zero_global_mean_moisture_advection=True,
        moisture_budget_correction=terms_to_modify,
        force_positive_names=force_positive_names,
    )

    mean_advection = gridded_ops.area_weighted_mean(
        data["tendency_of_total_water_path_due_to_advection"].to(device)
    )
    assert (mean_advection.abs() > 0.0).all()

    stepper_config = SingleModuleStepperConfig(
        builder=ModuleSelector(
            type="prebuilt",
            config={
                "module": Multiply(1.5).to(device),
            },
        ),
        in_names=list(data.keys()),
        out_names=list(data.keys()),
        normalization=NormalizationConfig(
            means={key: 0.0 for key in data.keys()},
            stds={key: 1.0 for key in data.keys()},
        ),
        corrector=corrector_config,
    )
    stepper = stepper_config.get_stepper(dataset_info=dataset_info)
    time = xr.DataArray(
        [
            [
                cftime.DatetimeProlepticGregorian(
                    2000, 1, int(i * 6 // 24) + 1, i * 6 % 24
                )
                for i in range(n_forward_steps + 1)
            ]
            for _ in range(3)
        ],
        dims=["sample", "time"],
    )
    batch_data = BatchData.new_on_cpu(
        data=data,
        time=time,
    ).to_device()
    # run the stepper on the data
    with torch.no_grad():
        stepped = stepper.train_on_batch(
            data=batch_data,
            optimization=NullOptimization(),
            compute_derived_variables=compute_derived_in_train_on_batch,
        )
    if not compute_derived_in_train_on_batch:
        stepped = stepped.compute_derived_variables()
    # check that the budget residual is zero
    budget_residual = stepped.gen_data["total_water_path_budget_residual"]
    if global_only:
        budget_residual = gridded_ops.area_weighted_mean(budget_residual)
    budget_residual = budget_residual.cpu().numpy()
    if terms_to_modify is not None:
        if global_only:
            mean_axis: tuple[int, ...] = (0,)
        else:
            mean_axis = (0, 2, 3)
        # first assert on timeseries, easier to look at
        np.testing.assert_almost_equal(
            np.abs(budget_residual).mean(axis=mean_axis), 0.0, decimal=6
        )
        np.testing.assert_almost_equal(budget_residual, 0.0, decimal=5)

    # check there is no mean advection
    mean_advection = (
        gridded_ops.area_weighted_mean(
            stepped.gen_data["tendency_of_total_water_path_due_to_advection"]
        )
        .cpu()
        .numpy()
    )
    np.testing.assert_almost_equal(mean_advection[:, 1:], 0.0, decimal=6)

    # check that the dry air is conserved
    dry_air = (
        gridded_ops.area_weighted_mean(
            AtmosphereData(
                stepped.gen_data, vertical_coordinate
            ).surface_pressure_due_to_dry_air
        )
        .cpu()
        .numpy()
    )
    dry_air_nonconservation = np.abs(dry_air[:, 1:] - dry_air[:, :-1])
    np.testing.assert_almost_equal(dry_air_nonconservation, 0.0, decimal=3)

    # check that positive forcing is enforced
    if force_positive:
        for name in force_positive_names:
            assert stepped.gen_data[name][:, :, 1:].min() >= 0.0


def _get_stepper(
    in_names: list[str],
    out_names: list[str],
    ocean_config: OceanConfig | None = None,
    module_name: Literal["AddOne", "ChannelSum", "RepeatChannel"] = "AddOne",
    norm_mean: float = 0.0,
    **kwargs,
):
    if module_name == "AddOne":

        class AddOne(torch.nn.Module):
            def forward(self, x):
                return x + 1

        module_config = {"module": AddOne()}
    elif module_name == "ChannelSum":
        # convenient for testing stepper with more inputs than outputs
        class ChannelSum(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.last_input: torch.Tensor | None = None

            def forward(self, x):
                self.last_input = x
                return x.sum(dim=-3, keepdim=True)

        module_config = {"module": ChannelSum()}
    elif module_name == "RepeatChannel":
        # convenient for testing stepper with more outputs than inputs
        class RepeatChannel(torch.nn.Module):
            def forward(self, x):
                return x.repeat(1, 2, 1, 1)

        module_config = {"module": RepeatChannel()}

    all_names = list(set(in_names + out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config=module_config),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={n: norm_mean for n in all_names},
            stds={n: 1.0 for n in all_names},
        ),
        ocean=ocean_config,
        **kwargs,
    )
    dataset_info = get_dataset_info()
    return config.get_stepper(dataset_info=dataset_info)


def test_step():
    stepper = _get_stepper(["a", "b"], ["a", "b"])
    input_data = {x: torch.rand(3, 5, 5).to(DEVICE) for x in ["a", "b"]}

    output = stepper.step(input_data, {})

    torch.testing.assert_close(output["a"], input_data["a"] + 1)
    torch.testing.assert_close(output["b"], input_data["b"] + 1)


def test_step_with_diagnostic():
    stepper = _get_stepper(["a"], ["a", "c"], module_name="RepeatChannel")
    input_data = {"a": torch.rand(3, 5, 5).to(DEVICE)}
    output = stepper.step(input_data, {})
    torch.testing.assert_close(output["a"], input_data["a"])
    torch.testing.assert_close(output["c"], input_data["a"])


@pytest.mark.parametrize("residual_prediction", [False, True])
def test_step_with_forcing_and_diagnostic(residual_prediction):
    norm_mean = 2.0
    stepper = _get_stepper(
        ["a", "b"],
        ["a", "c"],
        norm_mean=norm_mean,
        residual_prediction=residual_prediction,
    )
    input_data = {x: torch.rand(3, 5, 5).to(DEVICE) for x in ["a", "b"]}
    output = stepper.step(input_data, {})
    if residual_prediction:
        expected_a_output = 2 * input_data["a"] + 1 - norm_mean
    else:
        expected_a_output = input_data["a"] + 1
    torch.testing.assert_close(output["a"], expected_a_output)
    assert "b" not in output
    assert "c" in output


def test_step_with_prescribed_ocean():
    stepper = _get_stepper(
        ["a", "b"], ["a", "b"], ocean_config=OceanConfig("a", "mask")
    )
    input_data = {x: torch.rand(3, 5, 5).to(DEVICE) for x in ["a", "b"]}
    ocean_data = {x: torch.rand(3, 5, 5).to(DEVICE) for x in ["a", "mask"]}
    output = stepper.step(input_data, ocean_data)
    expected_a_output = torch.where(
        torch.round(ocean_data["mask"]).to(int) == 1,
        ocean_data["a"],
        input_data["a"] + 1,
    )
    torch.testing.assert_close(output["a"], expected_a_output)
    torch.testing.assert_close(output["b"], input_data["b"] + 1)
    assert set(output) == {"a", "b"}


def get_data_for_predict(
    n_steps, forcing_names: list[str]
) -> tuple[PrognosticState, BatchData]:
    n_samples = 3
    input_data = BatchData.new_on_device(
        data={"a": torch.rand(n_samples, 1, 5, 5).to(DEVICE)},
        time=xr.DataArray(
            np.zeros((n_samples, 1)),
            dims=["sample", "time"],
        ),
    ).get_start(
        prognostic_names=["a"],
        n_ic_timesteps=1,
    )
    forcing_data = BatchData.new_on_device(
        data={
            name: torch.rand(3, n_steps + 1, 5, 5).to(DEVICE) for name in forcing_names
        },
        time=xr.DataArray(
            np.zeros((n_samples, n_steps + 1)),
            dims=["sample", "time"],
        ),
    )
    return input_data, forcing_data


def test_predict():
    stepper = _get_stepper(["a"], ["a"])
    n_steps = 3
    input_data, forcing_data = get_data_for_predict(n_steps, forcing_names=[])
    forcing_data.data = {}
    output, new_input_data = stepper.predict(input_data, forcing_data)
    xr.testing.assert_allclose(forcing_data.time[:, 1:], output.time)
    variable = "a"
    assert output.data[variable].size(dim=1) == n_steps
    torch.testing.assert_close(
        output.data[variable][:, -1],
        input_data.as_batch_data().data[variable][:, 0] + n_steps,
    )
    assert isinstance(new_input_data, PrognosticState)
    new_input_state = new_input_data.as_batch_data()
    assert isinstance(new_input_state, BatchData)
    torch.testing.assert_close(
        new_input_state.data[variable][:, 0], output.data[variable][:, -1]
    )
    assert new_input_state.time.equals(output.time[:, -1:])


def test_predict_with_forcing():
    stepper = _get_stepper(["a", "b"], ["a"], module_name="ChannelSum")
    n_steps = 3
    input_data, forcing_data = get_data_for_predict(n_steps, forcing_names=["b"])
    output, new_input_data = stepper.predict(input_data, forcing_data)
    assert "b" not in output.data
    assert output.data["a"].size(dim=1) == n_steps
    xr.testing.assert_allclose(forcing_data.time[:, 1:], output.time)
    torch.testing.assert_close(
        output.data["a"][:, 0],
        input_data.as_batch_data().data["a"][:, 0] + forcing_data.data["b"][:, 0],
    )
    assert isinstance(new_input_data, PrognosticState)
    new_input_state = new_input_data.as_batch_data()
    assert isinstance(new_input_state, BatchData)
    torch.testing.assert_close(new_input_state.data["a"][:, 0], output.data["a"][:, -1])
    assert "b" not in new_input_state.data
    for n in range(1, n_steps):
        expected_a_output = output.data["a"][:, n - 1] + forcing_data.data["b"][:, n]
        torch.testing.assert_close(output.data["a"][:, n], expected_a_output)
    xr.testing.assert_equal(output.time, forcing_data.time[:, 1:])
    assert new_input_state.time.equals(output.time[:, -1:])


def test_predict_with_ocean():
    stepper = _get_stepper(["a"], ["a"], ocean_config=OceanConfig("a", "mask"))
    n_steps = 3
    input_data, forcing_data = get_data_for_predict(
        n_steps, forcing_names=["a", "mask"]
    )
    output, new_input_data = stepper.predict(input_data, forcing_data)
    xr.testing.assert_allclose(forcing_data.time[:, 1:], output.time)
    assert "mask" not in output.data
    assert output.data["a"].size(dim=1) == n_steps
    for n in range(n_steps):
        previous_a = (
            input_data.as_batch_data().data["a"][:, 0]
            if n == 0
            else output.data["a"][:, n - 1]
        )
        expected_a_output = torch.where(
            torch.round(forcing_data.data["mask"][:, n + 1]).to(int) == 1,
            forcing_data.data["a"][:, n + 1],
            previous_a + 1,
        )
        torch.testing.assert_close(output.data["a"][:, n], expected_a_output)
    assert isinstance(new_input_data, PrognosticState)
    new_input_state = new_input_data.as_batch_data()
    assert isinstance(new_input_state, BatchData)
    torch.testing.assert_close(new_input_state.data["a"][:, 0], output.data["a"][:, -1])
    assert new_input_state.time.equals(output.time[:, -1:])


def test_next_step_forcing_names():
    stepper = _get_stepper(
        ["a", "b", "c"],
        ["a"],
        module_name="ChannelSum",
        next_step_forcing_names=["c"],
    )
    input_data, forcing_data = get_data_for_predict(n_steps=1, forcing_names=["b", "c"])
    stepper.predict(input_data, forcing_data)
    assert len(stepper.modules) == 1
    torch.testing.assert_close(
        stepper.modules[0].module.last_input[:, 1, :],
        forcing_data.data["b"][:, 0],
    )
    torch.testing.assert_close(
        stepper.modules[0].module.last_input[:, 2, :],
        forcing_data.data["c"][:, 1],
    )


def test_stepper_from_state_using_resnorm_has_correct_normalizer():
    # If originally configured with a residual normalizer, the
    # stepper loaded from state should have the appropriately combined
    # full field and residual values in its loss_normalizer
    torch.manual_seed(0)
    full_field_means = {"a": 0.0, "b": 0.0, "diagnostic": 0.0}
    full_field_stds = {"a": 1.0, "b": 1.0, "diagnostic": 1.0}
    # residual scalings might have diagnostic variables but the stepper
    # should detect which prognostic variables to use from the set
    residual_means = {"a": 1.0, "b": 1.0, "diagnostic": 1.0}
    residual_stds = {"a": 2.0, "b": 2.0, "diagnostic": 2.0}
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(
            type="SphericalFourierNeuralOperatorNet", config={"scale_factor": 1}
        ),
        in_names=["a", "b"],
        out_names=["a", "b", "diagnostic"],
        normalization=NormalizationConfig(means=full_field_means, stds=full_field_stds),
        residual_normalization=NormalizationConfig(
            means=residual_means, stds=residual_stds
        ),
    )
    dataset_info = get_dataset_info()
    orig_stepper = config.get_stepper(
        dataset_info=dataset_info,
    )
    stepper_from_state = Stepper.from_state(orig_stepper.get_state())

    for stepper in [orig_stepper, stepper_from_state]:
        assert stepper.loss_obj.normalizer.means == {
            **residual_means,
            "diagnostic": full_field_means["diagnostic"],
        }
        assert stepper.loss_obj.normalizer.stds == {
            **residual_stds,
            "diagnostic": full_field_stds["diagnostic"],
        }
        assert stepper.normalizer.means == full_field_means
        assert stepper.normalizer.stds == full_field_stds


def get_regression_stepper_and_data():
    in_names = ["a", "b"]
    out_names = ["b", "c"]
    n_forward_steps = 2
    n_samples = 3
    img_shape = (9, 18)
    device = get_device()

    all_names = list(set(in_names + out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(
            type="SphericalFourierNeuralOperatorNet",
            config=dataclasses.asdict(
                SphericalFourierNeuralOperatorBuilder(
                    embed_dim=16,
                    num_layers=2,
                )
            ),
        ),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={n: 0.1 for n in all_names},
            stds={n: 1.1 for n in all_names},
        ),
        ocean=None,
    )
    dataset_info = get_dataset_info(img_shape=img_shape)
    stepper = config.get_stepper(dataset_info=dataset_info)
    data = BatchData(
        data={
            "a": torch.randn(n_samples, n_forward_steps + 1, *img_shape).to(device),
            "b": torch.randn(n_samples, n_forward_steps + 1, *img_shape).to(device),
            "c": torch.randn(n_samples, n_forward_steps + 1, *img_shape).to(device),
        },
        time=xr.DataArray(
            np.zeros((n_samples, n_forward_steps + 1)),
            dims=["sample", "time"],
        ),
    )
    return stepper, data


@pytest.mark.parametrize("use_optimization", [True, False])
def test_stepper_train_on_batch_regression(use_optimization: bool):
    torch.manual_seed(0)
    stepper, data = get_regression_stepper_and_data()
    if use_optimization:
        optimization_config = OptimizationConfig(
            optimizer_type="Adam",
            lr=0.0001,
        )
        optimization: OptimizationABC = optimization_config.build(
            stepper.modules, max_epochs=1
        )
    else:
        optimization = NullOptimization()
    result1 = stepper.train_on_batch(data, optimization)
    result2 = stepper.train_on_batch(data, optimization)
    output_dict = get_train_outputs_tensor_dict(result1, result2)
    validate_tensor_dict(
        output_dict,
        os.path.join(
            DIR, f"testdata/stepper_train_on_batch_regression-{use_optimization}.pt"
        ),
    )


def test_stepper_predict_regression():
    torch.manual_seed(0)
    stepper, data = get_regression_stepper_and_data()
    initial_condition = data.get_start(
        prognostic_names=["b"],
        n_ic_timesteps=1,
    )
    output, next_state = stepper.predict(
        initial_condition, data, compute_derived_variables=True
    )
    output_dict = get_predict_output_tensor_dict(output, next_state)
    validate_tensor_dict(
        output_dict,
        os.path.join(DIR, f"testdata/stepper_predict_regression.pt"),
    )


def get_predict_output_tensor_dict(
    output: BatchData, next_state: PrognosticState
) -> dict[str, torch.Tensor]:
    return flatten_dict(
        {
            "output": output.data,
            "next_state": next_state.as_batch_data().data,
        }
    )


def get_train_outputs_tensor_dict(
    step_1: TrainOutput, step_2: TrainOutput
) -> dict[str, torch.Tensor]:
    return flatten_dict(
        {
            "step_1": _get_train_output_tensor_dict(step_1),
            "step_2": _get_train_output_tensor_dict(step_2),
        }
    )


def flatten_dict(
    d: Mapping[str, Mapping[str, torch.Tensor]],
) -> dict[str, torch.Tensor]:
    return_dict = {}
    for k, v in d.items():
        for k2, v2 in v.items():
            return_dict[f"{k}.{k2}"] = v2
    return return_dict


def _get_train_output_tensor_dict(data: TrainOutput) -> dict[str, torch.Tensor]:
    return_dict = {}
    for k, v in data.metrics.items():
        return_dict[f"metrics.{k}"] = v
    for k, v in data.gen_data.items():
        assert v.shape[1] == 1
        return_dict[f"gen_data.{k}"] = v
    for k, v in data.target_data.items():
        assert v.shape[1] == 1
        return_dict[f"target_data.{k}"] = v
    return return_dict


def test_set_train_eval():
    stepper = _get_stepper(["a"], ["a"])
    for module in stepper.modules:
        assert module.training
    stepper.set_eval()
    for module in stepper.modules:
        assert not module.training
    stepper.set_train()
    for module in stepper.modules:
        assert module.training


def test_from_state_backwards_compatibility():
    """Ensure that the stepper can be loaded from a legacy style state dict.

    The state here is a simplified version of the one used in the checkpoint
    at https://huggingface.co/allenai/ACE2-ERA5/blob/main/ace2_era5_ckpt.tar"""

    class TestModule(torch.nn.Module):
        def __init__(self, param_shapes: Iterable[tuple[int, ...]]):
            super().__init__()
            for i, shape in enumerate(param_shapes):
                setattr(self, f"param{i}", torch.nn.Parameter(torch.randn(shape)))

    @ModuleSelector.register("test")
    @dataclasses.dataclass
    class TestModuleBuilder(ModuleConfig):
        param_shapes: list[tuple[int, ...]]

        def build(self, n_in_channels, n_out_channels, img_shape):
            return TestModule(self.param_shapes)

        @classmethod
        def from_state(cls, state):
            return cls(state["param_shapes"])

        def get_state(self):
            return {"param_shapes": self.param_shapes}

    model_shape = (10, 5)
    normalizer_state = {
        "means": {"surface_temperature": 1.0, "DSWRFtoa": 3.0},
        "stds": {"surface_temperature": 1.0, "DSWRFtoa": 3.0},
    }
    state = {
        "normalizer": normalizer_state,
        "img_shape": torch.Size([2, 4]),
        "module": {"module.param0": torch.zeros(model_shape)},
        "area": torch.ones((2, 4)),
        "sigma_coordinates": {
            "ak": torch.tensor([0.0000, 5119.8950]),
            "bk": torch.tensor([0.0000, 0.0000]),
        },
        "encoded_timestep": 21600000000,
        "loss_normalizer": normalizer_state,
        "config": {
            "builder": {
                "type": "test",
                "config": {"param_shapes": [model_shape]},
            },
            "in_names": ["DSWRFtoa", "surface_temperature"],
            "out_names": ["surface_temperature"],
            "normalization": {
                "global_means_path": "/statsdata/centering.nc",
                "global_stds_path": "/statsdata/scaling-full-field.nc",
                "exclude_names": None,
                "means": {},
                "stds": {},
            },
            "ocean": {
                "surface_temperature_name": "surface_temperature",
                "ocean_fraction_name": "ocean_fraction",
                "interpolate": False,
                "slab": None,
            },
            "loss": {"type": "MSE"},
            "next_step_forcing_names": ["DSWRFtoa"],
            "parameter_init": {  # legacy parameter init with both exclude and frozen
                "weights_path": "ckpt",
                "exclude_parameters": [],
                "frozen_parameters": {"include": [], "exclude": ["*"]},
                "alpha": 0.0,
                "beta": 0.0,
            },
        },
    }
    stepper = Stepper.from_state(state)
    # ensure newly created Stepper results in a loadable state
    new_state = stepper.get_state()
    Stepper.from_state(new_state)



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import dataclasses
from typing import Literal, Optional, Tuple, Union, cast

import torch as th
import torch.nn as nn

from .healpix_activations import CappedGELUConfig, UpsamplingBlockConfig
from .healpix_layers import HEALPixLayer

# RECURRENT BLOCKS


@dataclasses.dataclass
class RecurrentBlockConfig:
    """
    Configuration for the recurrent block.

    Parameters:
        in_channels: Number of input channels, default is 3.
        kernel_size: Size of the kernel, default is 1.
        enable_nhwc: Flag to enable NHWC data format, default is False.
        enable_healpixpad: Flag to enable HEALPix padding, default is False.
        block_type: Type of recurrent block, either "ConvGRUBlock" or "ConvLSTMBlock",
        default is "ConvGRUBlock".
    """

    in_channels: int = 3
    kernel_size: int = 1
    enable_nhwc: bool = False
    enable_healpixpad: bool = False
    block_type: Literal["ConvGRUBlock", "ConvLSTMBlock"] = "ConvGRUBlock"

    def build(self) -> nn.Module:
        """
        Builds the recurrent block model.

        Returns:
            Recurrent block.
        """
        if self.block_type == "ConvGRUBlock":
            return ConvGRUBlock(
                in_channels=self.in_channels,
                kernel_size=self.kernel_size,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        elif self.block_type == "ConvLSTMBlock":
            return ConvLSTMBlock(
                in_channels=self.in_channels,
                kernel_size=self.kernel_size,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        else:
            raise ValueError(f"Unsupported block type: {self.block_type}")


@dataclasses.dataclass
class ConvBlockConfig:
    """
    Configuration for the convolutional block.

    Parameters:
        in_channels: Number of input channels, default is 3.
        out_channels: Number of output channels, default is 1.
        kernel_size: Size of the kernel, default is 3.
        dilation: Dilation rate, default is 1.
        n_layers: Number of layers, default is 1.
        upsampling: Upsampling factor for TransposedConvUpsample, default is 2.
        upscale_factor: Upscale factor for ConvNeXtBlock and SymmetricConvNeXtBlock,
            default is 4.
        latent_channels: Number of latent channels, default is None.
        activation: Activation configuration, default is None.
        enable_nhwc: Flag to enable NHWC data format, default is False.
        enable_healpixpad: Flag to enable HEALPix padding, default is False.
        block_type: Type of block, default is "BasicConvBlock".
    """

    in_channels: int = 3
    out_channels: int = 1
    kernel_size: int = 3
    dilation: int = 1
    n_layers: int = 1
    stride: int = 2
    upscale_factor: int = 4
    latent_channels: Optional[int] = None
    upsampling: Optional[UpsamplingBlockConfig] = None
    activation: Optional[CappedGELUConfig] = None
    enable_nhwc: bool = False
    enable_healpixpad: bool = False
    block_type: Literal[
        "BasicConvBlock",
        "ConvNeXtBlock",
        "SymmetricConvNeXtBlock",
        "ConvThenUpsample",
        "TransposedConvUpsample",
    ] = "BasicConvBlock"

    def build(self) -> nn.Module:
        """
        Builds the convolutional block model.

        Returns:
            Convolutional block model.
        """
        if self.block_type == "BasicConvBlock":
            return BasicConvBlock(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                kernel_size=self.kernel_size,
                dilation=self.dilation,
                n_layers=self.n_layers,
                latent_channels=self.latent_channels,
                activation=self.activation,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        elif self.block_type == "ConvNeXtBlock":
            if self.latent_channels is None:
                self.latent_channels = 1
            return ConvNeXtBlock(
                in_channels=self.in_channels,
                latent_channels=cast(int, self.latent_channels),
                out_channels=self.out_channels,
                kernel_size=self.kernel_size,
                dilation=self.dilation,
                upscale_factor=self.upscale_factor,
                activation=self.activation,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        elif self.block_type == "SymmetricConvNeXtBlock":
            if self.latent_channels is None:
                self.latent_channels = 1
            return SymmetricConvNeXtBlock(
                in_channels=self.in_channels,
                latent_channels=cast(int, self.latent_channels),
                out_channels=self.out_channels,
                kernel_size=self.kernel_size,
                dilation=self.dilation,
                upscale_factor=self.upscale_factor,
                activation=self.activation,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        elif self.block_type == "ConvThenUpsample":
            return ConvThenUpsample(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                stride=self.stride,
                upsampling=self.upsampling,
                activation=self.activation,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        elif self.block_type == "TransposedConvUpsample":
            return TransposedConvUpsample(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                upsampling=self.stride,
                activation=self.activation,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        else:
            raise ValueError(f"Unsupported block type: {self.block_type}")


class ConvGRUBlock(nn.Module):
    """Class that implements a Convolutional GRU.

    Code modified from:
    https://github.com/happyjin/ConvGRU-pytorch/blob/master/convGRU.py
    """

    def __init__(
        self,
        in_channels=3,
        kernel_size=1,
        enable_nhwc=False,
        enable_healpixpad=False,
    ):
        """
        Args:
            in_channels: The number of input channels.
            kernel_size: Size of the convolutional kernel.
            enable_nhwc: Enable nhwc format, passed to wrapper.
            enable_healpixpad: If HEALPixPadding should be enabled, passed to wrapper.
        """
        super().__init__()

        self.channels = in_channels
        self.conv_gates = HEALPixLayer(
            layer=th.nn.Conv2d,
            in_channels=in_channels + self.channels,
            out_channels=2 * self.channels,  # for update_gate, reset_gate respectively
            kernel_size=kernel_size,
            padding="same",
            enable_nhwc=enable_nhwc,
            enable_healpixpad=enable_healpixpad,
        )
        self.conv_can = HEALPixLayer(
            layer=th.nn.Conv2d,
            in_channels=in_channels + self.channels,
            out_channels=self.channels,  # for candidate neural memory
            kernel_size=kernel_size,
            padding="same",
            enable_nhwc=enable_nhwc,
            enable_healpixpad=enable_healpixpad,
        )
        self.h = th.zeros(1, 1, 1, 1)

    def forward(self, inputs):
        """Forward pass of the ConvGRUBlock.

        Args:
            inputs: Input to the forward pass.

        Returns:
            th.Tensor: Result of the forward pass.
        """
        if inputs.shape != self.h.shape:
            self.h = th.zeros_like(inputs)
        combined = th.cat([inputs, self.h], dim=1)
        combined_conv = self.conv_gates(combined)

        gamma, beta = th.split(combined_conv, self.channels, dim=1)
        reset_gate = th.sigmoid(gamma)
        update_gate = th.sigmoid(beta)

        combined = th.cat([inputs, reset_gate * self.h], dim=1)
        cc_cnm = self.conv_can(combined)
        cnm = th.tanh(cc_cnm)

        h_next = (1 - update_gate) * self.h + update_gate * cnm
        self.h = h_next

        return inputs + h_next

    def reset(self):
        """Reset the update gates."""
        self.h = th.zeros_like(self.h)


class ConvLSTMBlock(nn.Module):
    """Convolutional LSTM block."""

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 1,
        latent_channels: int = 1,
        kernel_size: int = 3,
        downscale_factor: int = 4,
        upscale_factor: int = 4,
        n_layers: int = 1,
        latent_conv_size: int = 3,  # Add latent_conv_size parameter
        dilation: int = 1,
        activation: nn.Module = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Args:
            in_channels: The number of input channels.
            out_channels: The number of output channels.
            latent_channels: Number of latent channels.
            kernel_size: Size of the convolutional kernel.
            downscale_factor: Downscale factor.
            upscale_factor: Upscale factor.
            n_layers: Number of layers.
            latent_conv_size: Size of latent convolution.
            dilation: Spacing between kernel points.
            activation: Activation function.
            enable_nhwc: Enable nhwc format.
            enable_healpixpad: If HEALPixPadding should be enabled.
        """
        super().__init__()
        # Instantiate 1x1 conv to increase/decrease channel depth if necessary
        # Skip connection for output
        if in_channels == out_channels:
            self.skip_module = lambda x: x  # Identity-function required in forward pass
        else:
            self.skip_module = HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=in_channels,  # out channels describes the space of the output of conv here; but we have the output of LSTM which is the input layer size
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        # Convolution block
        convblock = []
        # 3x3 convolution increasing channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels
                * 2,  # accounts for the h layer, which is concatenated before convolution runs
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation)
        # 3x3 convolution maintaining increased channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation)

        # Now for the LSTM bit
        self.channels = in_channels
        self.lstm_gates = HEALPixLayer(
            layer=th.nn.Conv2d,
            in_channels=latent_channels * upscale_factor,
            out_channels=self.channels
            * 4,  # for input_gate, forget_gate, cell_gate, output_gate respectively (LSTM)
            kernel_size=kernel_size,
            padding="same",
            enable_nhwc=enable_nhwc,
            enable_healpixpad=enable_healpixpad,
        )
        self.h = th.zeros(1, 1, 1, 1)
        self.c = th.zeros(1, 1, 1, 1)
        self.convblock = nn.Sequential(*convblock)

    def forward(self, inputs):
        """Forward pass of the ConvLSTMBlock.

        Args:
            x: Inputs to the forward pass.

        Returns:
            th.Tensor: Result of the forward pass.
        """
        if inputs.shape != self.h.shape:
            self.h = th.zeros_like(inputs)
            self.c = th.zeros_like(inputs)

        combined = th.cat([inputs, self.h], dim=1)
        conv_outputs = self.convblock(combined)

        lstm_gates = self.lstm_gates(conv_outputs)

        # Split the combined_conv into input_gate, forget_gate, cell_gate, output_gate
        i, f, c_hat, o = th.split(lstm_gates, self.channels, dim=1)
        input_gate = th.sigmoid(i)
        forget_gate = th.sigmoid(f)
        cell_gate = th.tanh(c_hat)
        output_gate = th.sigmoid(o)

        self.c = forget_gate * self.c + input_gate * cell_gate
        self.h = output_gate * th.tanh(self.c)

        skip_connection = self.skip_module(inputs)
        return skip_connection + self.h

    def reset(self):
        self.h = th.zeros_like(self.h)
        self.c = th.zeros_like(self.c)


# CONV BLOCKS


class BasicConvBlock(nn.Module):
    """Convolution block consisting of n subsequent convolutions and activations."""

    def __init__(
        self,
        in_channels=3,
        out_channels=1,
        kernel_size=3,
        dilation=1,
        n_layers=1,
        latent_channels=None,
        activation=None,
        enable_nhwc=False,
        enable_healpixpad=False,
    ):
        """
        Args:
            in_channels: The number of input channels.
            out_channels: The number of output channels.
            kernel_size: Size of the convolutional kernel.
            dilation: Spacing between kernel points, passed to nn.Conv2d.
            n_layers: Number of convolutional layers.
            latent_channels: Number of latent channels.
            activation: ModuleConfig for activation function to use.
            enable_nhwc: Enable nhwc format, passed to wrapper.
            enable_healpixpad:: If HEALPixPadding should be enabled, passed to wrapper.
        """
        super().__init__()
        if latent_channels is None:
            latent_channels = max(in_channels, out_channels)
        convblock = []
        for n in range(n_layers):
            convblock.append(
                HEALPixLayer(
                    layer=th.nn.Conv2d,
                    in_channels=in_channels if n == 0 else latent_channels,
                    out_channels=out_channels if n == n_layers - 1 else latent_channels,
                    kernel_size=kernel_size,
                    dilation=dilation,
                    enable_nhwc=enable_nhwc,
                    enable_healpixpad=enable_healpixpad,
                )
            )
            if activation is not None:
                convblock.append(activation.build())
        self.convblock = nn.Sequential(*convblock)

    def forward(self, x):
        """Forward pass of the BasicConvBlock.

        Args:
            x: Inputs to the forward pass.

        Returns:
            th.Tensor: Result of the forward pass.
        """
        return self.convblock(x)


class ConvNeXtBlock(nn.Module):
    """A modified ConvNeXt network block as described in the paper
    "A ConvNet for the 21st Century" (https://arxiv.org/pdf/2201.03545.pdf).

    This block consists of a series of convolutional layers with optional activation functions,
    and a residual connection.

    Parameters:
        skip_module: A module to align the input and output channels for the residual connection.
        convblock: A sequential container of convolutional layers with optional activation functions.
    """

    def __init__(
        self,
        in_channels: int = 3,
        latent_channels: int = 1,
        out_channels: int = 1,
        kernel_size: int = 3,
        dilation: int = 1,
        upscale_factor: int = 4,
        activation: Optional[CappedGELUConfig] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Initializes a ConvNeXtBlock instance with specified parameters.

        Args:
            in_channels: Number of input channels.
            latent_channels: Number of latent channels used in the block.
            out_channels: Number of output channels.
            kernel_size: Size of the convolutional kernels.
            dilation: Dilation rate for convolutions.
            upscale_factor: Factor by which to upscale the number of latent channels.
            activation: Configuration for the activation function used between layers.
            enable_nhwc: Whether to enable NHWC format.
            enable_healpixpad: Whether to enable HEALPixPadding.
        """
        super().__init__()

        # Instantiate 1x1 conv to increase/decrease channel depth if necessary
        if in_channels == out_channels:
            self.skip_module = lambda x: x  # Identity-function required in forward pass
        else:
            self.skip_module = HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        # Convolution block
        convblock = []
        # 3x3 convolution increasing channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        # 3x3 convolution maintaining increased channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        # Linear postprocessing
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=out_channels,
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        self.convblock = nn.Sequential(*convblock)

    def forward(self, x):
        """Forward pass of the ConvNeXtBlock.

        Args:
            x: Input tensor.

        Returns:
            The result of the forward pass.
        """
        return self.skip_module(x) + self.convblock(x)


class DoubleConvNeXtBlock(nn.Module):
    """A variant of the ConvNeXt block that includes two sequential ConvNeXt blocks within a single module.

    Parameters:
        skip_module1: A module to align the input and intermediate channels for the first residual connection.
        skip_module2: A module to align the intermediate and output channels for the second residual connection.
        convblock1: A sequential container of convolutional layers for the first ConvNeXt block.
        convblock2: A sequential container of convolutional layers for the second ConvNeXt block.
    """

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 1,
        kernel_size: int = 3,
        dilation: int = 1,
        upscale_factor: int = 4,
        latent_channels: int = 1,
        activation: Optional[CappedGELUConfig] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Initializes a DoubleConvNeXtBlock instance with specified parameters.

        Args:
            in_channels: Number of input channels (default is 3).
            out_channels: Number of output channels (default is 1).
            kernel_size: Size of the convolutional kernels (default is 3).
            dilation: Dilation rate for convolutions (default is 1).
            upscale_factor: Factor by which to upscale the number of latent channels (default is 4).
            latent_channels: Number of latent channels used in the block (default is 1).
            activation: Configuration for the activation function used between layers (default is None).
            enable_nhwc: Whether to enable NHWC format (default is False).
            enable_healpixpad: Whether to enable HEALPixPadding (default is False).
        """
        super().__init__()

        if in_channels == int(latent_channels):
            self.skip_module1 = (
                lambda x: x
            )  # Identity-function required in forward pass
        else:
            self.skip_module1 = HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=int(latent_channels),
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        if out_channels == int(latent_channels):
            self.skip_module2 = (
                lambda x: x
            )  # Identity-function required in forward pass
        else:
            self.skip_module2 = HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=out_channels,
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )

        # 1st ConvNeXt block, the output of this one remains internal
        convblock1 = []
        # 3x3 convolution establishing latent channels channels
        convblock1.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=int(latent_channels),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock1.append(activation.build())
        # 1x1 convolution establishing increased channels
        convblock1.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock1.append(activation.build())
        # 1x1 convolution returning to latent channels
        convblock1.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=int(latent_channels),
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock1.append(activation.build())
        self.convblock1 = nn.Sequential(*convblock1)

        # 2nd ConNeXt block, takes the output of the first convnext block
        convblock2 = []
        # 3x3 convolution establishing latent channels channels
        convblock2.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=int(latent_channels),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock2.append(activation.build())
        # 1x1 convolution establishing increased channels
        convblock2.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock2.append(activation.build())
        # 1x1 convolution reducing to output channels
        convblock2.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=out_channels,
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock2.append(activation.build())
        self.convblock2 = nn.Sequential(*convblock2)

    def forward(self, x):
        """Forward pass of the DoubleConvNextBlock
        Args:
            x: inputs to the forward pass
        Returns:
            result of the forward pass
        """
        # internal convnext result
        x1 = self.skip_module1(x) + self.convblock1(x)
        # return second convnext result
        return self.skip_module2(x1) + self.convblock2(x1)


class SymmetricConvNeXtBlock(nn.Module):
    """A symmetric variant of the ConvNeXt block, with convolutional layers mirrored
    around a central axis for symmetric feature extraction.

    Parameters:
        skip_module1: A module to align the input and intermediate channels for the first residual connection.
        skip_module2: A module to align the intermediate and output channels for the second residual connection.
        convblock1: A sequential container of convolutional layers for the symmetric ConvNeXt block.
    """

    def __init__(
        self,
        in_channels: int = 3,
        latent_channels: int = 1,
        out_channels: int = 1,
        kernel_size: int = 3,
        dilation: int = 1,
        upscale_factor: int = 4,
        activation: Optional[CappedGELUConfig] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Initializes a SymmetricConvNeXtBlock instance with specified parameters.

        Args:
            in_channels: Number of input channels (default is 3).
            out_channels: Number of output channels (default is 1).
            kernel_size: Size of the convolutional kernels (default is 3).
            dilation: Dilation rate for convolutions (default is 1).
            upscale_factor: Upscale factor.
            latent_channels: Number of latent channels used in the block (default is 1).
            activation: Configuration for the activation function used between layers (default is None).
            enable_nhwc: Whether to enable NHWC format (default is False).
            enable_healpixpad: Whether to enable HEALPixPadding (default is False).
        """
        if in_channels == int(latent_channels):
            self.skip_module = lambda x: x  # Identity-function required in forward pass
        else:
            self.skip_module = HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )

        # 1st ConvNeXt block, the output of this one remains internal
        convblock = []
        # 3x3 convolution establishing latent channels channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=in_channels,
                out_channels=int(latent_channels),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        # 1x1 convolution establishing increased channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=int(latent_channels * upscale_factor),
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        # 1x1 convolution returning to latent channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels * upscale_factor),
                out_channels=int(latent_channels),
                kernel_size=1,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        # 3x3 convolution from latent channels to latent channels
        convblock.append(
            HEALPixLayer(
                layer=th.nn.Conv2d,
                in_channels=int(latent_channels),
                out_channels=out_channels,  # int(latent_channels),
                kernel_size=kernel_size,
                dilation=dilation,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            convblock.append(activation.build())
        self.convblock = nn.Sequential(*convblock)

    def forward(self, x):
        """Forward pass of the SymmetricConvNextBlock
        Args:
            x: inputs to the forward pass
        Returns:
            result of the forward pass
        """
        # residual connection with reshaped inpute and output of conv block
        return self.skip_module(x) + self.convblock(x)


class ConvThenUpsample(nn.Module):
    """Wrapper for upsampling and then applying a convolution using HEALPix or other tensor data.
    Allows more control over the type of upsampling (smooth with bilinear or pixelated
    with nearest-neighbor) and feature extraction.
    This class wraps the `nn.Upsample` and `nn.Conv2d` classes to replace ConvTranspose2d and handle tensor data with
    HEALPix or other geometry layers.
    """

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 1,
        stride: int = 2,
        upsampling: Optional[UpsamplingBlockConfig] = None,
        activation: Optional[CappedGELUConfig] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        super().__init__()
        upsampler = []
        if upsampling is not None:
            upsampler.append(upsampling.build())
        # Upsample transpose conv
        upsampler.append(
            HEALPixLayer(
                layer=nn.Conv2d,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=stride,
                stride=stride,
                padding=0,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            upsampler.append(activation.build())
        self.upsampler = nn.Sequential(*upsampler)

    def forward(self, x):
        """Forward pass of the ConvThenUpsample layer.
        Args:
            x: The values to upsample.
        Returns:
            th.Tensor: The upsampled values.
        """
        return self.upsampler(x)


class TransposedConvUpsample(nn.Module):
    """Wrapper for upsampling with a transposed convolution using HEALPix or other tensor data.

    This class wraps the `nn.ConvTranspose2d` class to handle tensor data with
    HEALPix or other geometry layers.
    """

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 1,
        upsampling: int = 2,
        activation: Optional[CappedGELUConfig] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Args:
            in_channels: The number of input channels.
            out_channels: The number of output channels.
            upsampling: Stride size that will be used for upsampling.
            activation: ModuleConfig for the activation function used in upsampling.
            enable_nhwc: Enable nhwc format, passed to wrapper.
            enable_healpixpad: If HEALPixPadding should be enabled, passed to wrapper.
        """
        super().__init__()
        upsampler = []
        # Upsample transpose conv
        upsampler.append(
            HEALPixLayer(
                layer=nn.ConvTranspose2d,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=upsampling,
                stride=upsampling,
                padding=0,
                enable_nhwc=enable_nhwc,
                enable_healpixpad=enable_healpixpad,
            )
        )
        if activation is not None:
            upsampler.append(activation.build())
        self.upsampler = nn.Sequential(*upsampler)

    def forward(self, x):
        """Forward pass of the TransposedConvUpsample layer.

        Args:
            x: The values to upsample.

        Returns:
            th.Tensor: The upsampled values.
        """
        return self.upsampler(x)


# Helpers


class Interpolate(nn.Module):
    """Helper class for interpolation.

    This class handles interpolation, storing scale factor and mode for
    `nn.functional.interpolate`.
    """

    def __init__(self, scale_factor: Union[int, Tuple], mode: str = "nearest"):
        """
        Args:
            scale_factor: Multiplier for spatial size, passed to `nn.functional.interpolate`.
            mode, : Interpolation mode used for upsampling, passed to `nn.functional.interpolate`.
        """
        super().__init__()
        self.interp = nn.functional.interpolate
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, inputs):
        """Forward pass of the Interpolate layer.

        Args:
            inputs: Inputs to interpolate.

        Returns:
            th.Tensor: The interpolated values.
        """
        return self.interp(inputs, scale_factor=self.scale_factor, mode=self.mode)



import torch

import fme
from fme.ace.data_loading.perturbation import (
    ConstantConfig,
    GreensFunctionConfig,
    PerturbationSelector,
)


def test_constant_perturbation_config():
    selector = PerturbationSelector(
        type="constant",
        config={"amplitude": 1.0},
    )
    perturbation = selector.build()
    assert isinstance(perturbation, ConstantConfig)
    assert perturbation.amplitude == 1.0
    nx, ny = 5, 5
    lat = torch.arange(nx, device=fme.get_device())
    lon = torch.arange(ny, device=fme.get_device())
    lats, lons = torch.meshgrid(lat, lon, indexing="ij")
    ocean_fraction = torch.ones(nx, ny, device=fme.get_device())
    data = torch.ones(nx, ny, device=fme.get_device())
    expected = 2.0 * torch.ones(nx, ny, device=fme.get_device())
    perturbation.apply_perturbation(data, lats, lons, ocean_fraction)
    torch.testing.assert_close(data, expected)


def test_green_function_perturbation_config():
    selector = PerturbationSelector(
        type="greens_function",
        config={
            "amplitude": 1.0,
            "lat_center": 0.0,
            "lon_center": 0.0,
            "lat_width": 10.0,
            "lon_width": 10.0,
        },
    )
    perturbation = selector.build()
    assert isinstance(perturbation, GreensFunctionConfig)
    assert perturbation.amplitude == 1.0
    assert perturbation.lat_center == 0.0
    assert perturbation.lon_center == 0.0
    assert perturbation.lat_width == 10.0
    assert perturbation.lon_width == 10.0
    nx, ny = 5, 5
    lat = torch.arange(nx, device=fme.get_device())
    lon = torch.arange(ny, device=fme.get_device())
    lats, lons = torch.meshgrid(lat, lon, indexing="ij")
    ocean_fraction = torch.ones(nx, ny, device=fme.get_device())
    data = torch.ones(nx, ny, device=fme.get_device())
    perturbation.apply_perturbation(data, lats, lons, ocean_fraction)



"""
Utils for CM4 ocean data preprocessing. This script relies on a fork of the
m2lines/ocean_emulators repo:
https://github.com/jpdunc23/ocean_emulators/tree/cm4-preprocessing.

The 200-year pre-industrial control simulation ocean preprocessing ran in about
6 hours on LEAP's 2i2c JupyterHub in the default "notebook" conda environment
using Dask Gateway with 64x '8CPU, 57.9Gi' workers. See the
CM4-pre-industrial-control-ocean-200yr.yaml and this notebook for details:
https://github.com/ai2cm/explore/blob/master/elynn/2024-11-11-CM4-200yr-preprocessing-ocean.ipynb
"""

import dataclasses
import os
import pdb
import sys
from typing import IO, Any, Mapping, Optional, Protocol, Sequence, Tuple, Union

import click
import dacite
import fsspec
import numpy as np
import xarray as xr
import xpartition  # noqa
import yaml
from compute_dataset import ChunkingConfig, StandardDimMapping
from dask.diagnostics import ProgressBar
from dask.distributed import Client, LocalCluster
from get_stats import StatsConfig
from ocean_emulators.preprocessing import (
    horizontal_regrid,
    rotate_vectors,
    spatially_filter,
)
from ocean_emulators.simulation_preprocessing.gfdl_cm4 import cm4_preprocessing


@dataclasses.dataclass
class OceanStandardNameMapping(StandardDimMapping):
    longitude_dim: str = "x"
    latitude_dim: str = "y"
    time_dim: str = "time"
    vertical_dim: str = "lev"
    vertical_idim: str = "ilev"
    rotation_angle: str = "angle"
    sea_water_x_velocity: str = "uo"
    sea_water_y_velocity: str = "vo"
    sea_water_salinity: str = "so"
    sea_water_potential_temperature: str = "thetao"
    surface_temperature: str = "tos"
    surface_downward_x_stress: str = "tauuo"
    surface_downward_y_stress: str = "tauvo"
    sea_ice_x_velocity: str = "UI"
    sea_ice_y_velocity: str = "VI"
    sea_ice_modeled: str = "EXT"
    sea_ice_fraction: str = "sea_ice_fraction"
    wetmask: str = "wetmask"
    ocean_layer_thickness: str = "layer_thickness"

    def __post_init__(self):
        self.full_field_dims = [self.longitude_dim, self.latitude_dim, self.time_dim]

    @property
    def rotated_vars(self) -> Sequence[Tuple[str, str]]:
        return (
            (self.sea_water_x_velocity, self.sea_water_y_velocity),
            (self.sea_ice_x_velocity, self.sea_ice_y_velocity),
            (self.surface_downward_x_stress, self.surface_downward_y_stress),
        )

    @property
    def vars_3d(self) -> Sequence[str]:
        return (
            self.sea_water_x_velocity,
            self.sea_water_y_velocity,
            self.sea_water_salinity,
            self.sea_water_potential_temperature,
        )


def _rename(ds: xr.Dataset, renaming: Mapping[str, str]):
    if len(renaming) == 0:
        return ds
    return ds.rename(renaming)


@dataclasses.dataclass
class CoarseningConfig:
    """Configuration for coarsening a supplemental dataset with higher temporal
    frequency. Assumed to have the same horizontal coordinates as the other
    datasets.

    Attributes:
        zarr: name of zarr with the higher frequency data.
        n_coarsen: number of timepoints over which to take the mean.
        renaming: (optional) mapping of names in dataset to renamed output.

    """

    zarr: str
    n_coarsen: int
    renaming: Mapping[str, str] = dataclasses.field(default_factory=dict)

    def coarsen(self, ds: xr.Dataset, time_dim: str):
        return ds.coarsen({time_dim: self.n_coarsen}).mean().drop_vars(time_dim)

    def rename(self, ds: xr.Dataset):
        return _rename(ds, self.renaming)


@dataclasses.dataclass
class OceanStaticConfig:
    """Configuration for static ocean data.

    Attributes:
        zarr: name of zarr with the static ocean data.
        names: names of variables to extract from the zarr.
        renaming: (optional) mapping of names in dataset to renamed output.


    """

    zarr: str
    names: Sequence[str]
    renaming: Mapping[str, str] = dataclasses.field(default_factory=dict)

    def rename(self, ds: xr.Dataset):
        return _rename(ds, self.renaming)


@dataclasses.dataclass
class DaskConfig:
    """Configuration for Dask, either LocalCluster or dask-gateway. See
    https://docs.2i2c.org/user/howto/launch-dask-gateway-cluster for
    dask-gateway usage on the LEAP-Pangeo 2i2c hub, where the relevant option is
    "worker_resource_allocation" which can be one of "1CPU, 7.2Gi", "2CPU, 14.5Gi",
    "4CPU, 28.9Gi", "8CPU, 57.9Gi", and "16CPU, 115.8Gi". Prepend
    https://leap.2i2c.cloud to dask dashboard urls when using the LEAP-Pangeo
    2i2c hub.

    Attributes:
        n_workers: number of Dask workers.
        use_gateway: whether to use dask-gateway
        cluster_options: additional options for configuring the LocalCluster or
            Gateway.

    """

    n_workers: int = 16
    use_gateway: bool = False
    cluster_options: Mapping[str, Any] = dataclasses.field(default_factory=dict)
    _cluster = None

    def get_client(self) -> Client:
        if self._cluster is not None:
            return self._cluster.get_client()
        if self.use_gateway:
            from dask_gateway import Gateway

            # use default gateway settings
            gateway = Gateway()
            options = gateway.cluster_options()
            options.update(self.cluster_options)
            self._cluster = gateway.new_cluster(options)
            self._cluster.scale(self.n_workers)
        else:
            self._cluster = LocalCluster(
                n_workers=self.n_workers, **self.cluster_options
            )
        return self._cluster.get_client()

    def close_cluster(self):
        if self._cluster is not None:
            self._cluster.close()
            self._cluster = None


@dataclasses.dataclass
class OceanDatasetComputationConfig:
    """Configuration of computation details for an FME reference dataset.

    Attributes:
        ocean_zarr: name of zarr with ocean model outputs.
        ice_zarr: name of zarr with ice model outputs.
        nc_grid_path: path to nc file with variables needed by
            ocean_emulators.simulation_processing.gfdl_om4.om4_preprocessing.
        nc_mosaic_path: path to nc file for grid mosaic with variables needed by
            ocean_emulators.simulation_processing.gfdl_om4.convert_super_grid.
        nc_target_grid_path: path to nc file for target grid.
        ocean_static_zarr: (optional) name of zarr with static ocean data.
        renaming: (optional) mapping of names in dataset to renamed output.
        standard_names: (optional) mapping of standard names to corresponding
            names of variables in the dataset.
        chunking: (optional) config for dask chunking.
    """

    ocean_zarr: str
    ice_zarr: str
    nc_grid_path: str
    nc_mosaic_path: str
    nc_target_grid_path: str
    coarsen: Optional[CoarseningConfig] = None
    ocean_static: Optional[OceanStaticConfig] = None
    renaming: Mapping[str, str] = dataclasses.field(default_factory=dict)
    standard_names: OceanStandardNameMapping = dataclasses.field(
        default_factory=OceanStandardNameMapping
    )
    chunking: ChunkingConfig = dataclasses.field(default_factory=ChunkingConfig)
    ocean_dataset_nc_files: str = ""
    ocean_dataset_monthly_layer_thickness_files: str = ""
    compute_e3sm_surface_downward_heat_flux: bool = False
    ice_dataset_nc_files: str = ""
    ocean_vertical_target_interface_levels: list = dataclasses.field(
        default_factory=list
    )

    def rename(self, ds: xr.Dataset):
        return _rename(ds, self.renaming)


class FileSystemProtocol(Protocol):
    def get_mapper(self, url: str) -> Union[str, fsspec.mapping.FSMap]: ...

    def open(self, *args, **kwargs) -> IO: ...


class _DummyFileSystem(FileSystemProtocol):
    def get_mapper(self, url: str) -> str:
        return url

    def open(self, *args, **kwargs) -> IO:
        return open(*args, **kwargs)


_DUMMY_FS = _DummyFileSystem()


@dataclasses.dataclass
class FileSystemConfig(FileSystemProtocol):
    """Configuration for creating an fsspec.filesystem.

    Attributes:
        protocol: file system protocol.
        fs_options: options for file system.

    """

    protocol: str
    storage_options: Mapping[str, Any]

    def __post_init__(self):
        self._fs = fsspec.filesystem(self.protocol, **self.storage_options)

    def get_mapper(self, url: str) -> fsspec.mapping.FSMap:
        return self._fs.get_mapper(url)


@dataclasses.dataclass
class OceanDatasetConfig:
    """Dataset provenance for a set of reference simulations.

    Attributes:
        runs: mapping of short names to full paths of reference datasets.
        data_output_directory: path to parent directory where the output zarr
            will be written.
        dataset_computation: configuration details for compute_lazy_dataset.
        stats_config: configuration to retrieve statistics dataset.
        n_split: number of xpartition partitions to use when writing the data.
        dask: (optional) configuration for the dask cluster.
        filesystem: (optional) file system to use for reading and writing data.

    """

    runs: Mapping[str, str]
    data_output_directory: str
    dataset_computation: OceanDatasetComputationConfig
    stats: StatsConfig
    n_split: int = 10
    dask: Optional[DaskConfig] = None
    filesystem: Optional[FileSystemConfig] = None

    @classmethod
    def from_file(cls, path: str) -> "OceanDatasetConfig":
        with open(path, "r") as file:
            data = yaml.safe_load(file)

        return dacite.from_dict(
            data_class=cls, data=data, config=dacite.Config(cast=[tuple], strict=True)
        )


def compute_lazy_dataset(
    config: OceanDatasetComputationConfig,
    run_directory: str,
    fs: Optional[FileSystemProtocol] = None,
    backend_kwargs=None,
):
    """Compute CM4 ocean and ice dataset.

    Arguments:
        config: Path to dataset configuration YAML file.
        run_directory: Path to reference run directory.
    """
    if fs is None:
        fs = _DUMMY_FS

    xr.set_options(keep_attrs=True)

    ocean_names = config.standard_names
    time_dim = ocean_names.time_dim
    lat_dim = ocean_names.latitude_dim
    lon_dim = ocean_names.longitude_dim
    vdim = ocean_names.vertical_dim
    vidim = ocean_names.vertical_idim

    om_zarr_path = os.path.join(run_directory, config.ocean_zarr)
    sis_zarr_path = os.path.join(run_directory, config.ice_zarr)

    ds = cm4_preprocessing(
        om_zarr_path=om_zarr_path,
        sis_zarr_path=sis_zarr_path,
        nc_grid_path=config.nc_grid_path,
        nc_mosaic_path=config.nc_mosaic_path,
        fs=fs,
        backend_kwargs=backend_kwargs,
    )
    ds = ds.chunk(
        {
            vdim: 1,
            lat_dim: -1,
            lon_dim: -1,
        }
    )

    urls = [
        om_zarr_path,
        sis_zarr_path,
        config.nc_grid_path,
        config.nc_mosaic_path,
        config.nc_target_grid_path,
    ]

    if config.coarsen is not None:
        zarr_path = fs.get_mapper(os.path.join(run_directory, config.coarsen.zarr))
        urls.append(zarr_path)
        ds_coarsen = xr.open_zarr(zarr_path)
        ds_coarsen = config.coarsen.coarsen(ds_coarsen, time_dim)
        ds_coarsen = config.coarsen.rename(ds_coarsen)
        ds_coarsen = ds_coarsen.chunk(
            {time_dim: ds.chunks[time_dim], lon_dim: -1, lat_dim: -1}
        )
        ds = xr.merge([ds, ds_coarsen])

    if config.ocean_static is not None:
        zarr_path = os.path.join(run_directory, config.ocean_static.zarr)
        urls.append(zarr_path)
        with fs.open(zarr_path) as f:
            ds_static = xr.open_dataset(
                zarr_path,
                decode_timedelta=False,
                engine="zarr",
                backend_kwargs=backend_kwargs,
            )[config.ocean_static.names]
        ds_static = config.ocean_static.rename(ds_static)
        ds = xr.merge([ds, ds_static])

    idepth_data = {}

    for i, depth in enumerate(ds[vidim].values):
        idepth_data[f"idepth_{i}"] = xr.DataArray(depth)
        idepth_data[f"idepth_{i}"].attrs["units"] = "meters"
        idepth_data[f"idepth_{i}"].attrs["long_name"] = f"Depth at interface level-{i}"

    idepth_ds = xr.Dataset(idepth_data)

    assert (
        ocean_names.sea_ice_fraction in ds
    ), f"Sea ice fraction variable {ocean_names.sea_ice_fraction} is missing."

    print(f"Preprocessed size: {ds.nbytes / 1e9:.1f} GB")

    # save attributes to add back after processing
    attrs = {}
    for var in ds.data_vars:
        attrs[var] = ds[var].attrs

    angle = ds[ocean_names.rotation_angle]
    for varname_x, varname_y in ocean_names.rotated_vars:
        x_rotated, y_rotated = rotate_vectors(ds[varname_x], ds[varname_y], angle)
        ds[varname_x] = x_rotated.astype(np.float32)
        ds[varname_y] = y_rotated.astype(np.float32)

    # spatial filtering
    ds = spatially_filter(
        ds, ds[ocean_names.wetmask], depth_dim=vdim, y_dim=lat_dim, x_dim=lon_dim
    )

    # regrid
    with fs.open(config.nc_target_grid_path) as f:
        ds_target_grid = xr.open_dataset(f).load()

    # TODO: remove target grid dimension assumptions
    ds_target_grid = ds_target_grid.rename(
        {
            "grid_x": "x_b",
            "grid_y": "y_b",
            "grid_xt": lon_dim,
            "grid_yt": lat_dim,
            "grid_lon": "lon_b",
            "grid_lat": "lat_b",
            "grid_lont": "lon",
            "grid_latt": "lat",
        }
    )
    # fill nans in sea_ice_fraction to be
    # consistent with ocean fraction in ocean_emulators
    if ocean_names.sea_ice_fraction in ds.data_vars:
        ds[ocean_names.sea_ice_fraction] = ds[ocean_names.sea_ice_fraction].fillna(0.0)
    ds_regridded = horizontal_regrid(ds, ds_target_grid).astype("float32")
    if ocean_names.sea_ice_fraction in ds_regridded.data_vars:
        ds_regridded[ocean_names.sea_ice_fraction] = ds_regridded[
            ocean_names.sea_ice_fraction
        ].fillna(0.0)
    print(f"Regridded size: {ds_regridded.nbytes / 1e9:.1f} GB")

    ds = ds_regridded
    for var, attrs in attrs.items():
        ds[var].attrs = attrs

    # fill ice velocity with NaN where sea ice is 0
    cond = ds[ocean_names.sea_ice_modeled] > 0.0

    for var in [ocean_names.sea_ice_x_velocity, ocean_names.sea_ice_y_velocity]:
        ds[var] = ds[var].where(cond, np.nan)

    wetmask = ds[ocean_names.wetmask].astype(np.float32)
    if len(wetmask.attrs) == 0:
        wetmask.attrs["long_name"] = "ocean mask"
        wetmask.attrs["units"] = "0 if land, 1 if ocean"

    ds["mask"] = wetmask
    vars_3d = list(ocean_names.vars_3d) + ["mask"]

    for i, _ in enumerate(ds[vdim].values):
        for var in vars_3d:
            long_name = ds[var].long_name
            ds[f"{var}_{i}"] = ds[var].isel({vdim: i})
            ds[f"{var}_{i}"].attrs["long_name"] = long_name + f" level-{i}"

    ds = ds.drop_vars(vars_3d)
    ds = ds.drop_dims(vdim)
    ds = ds.reset_coords(drop=True)
    ds = xr.merge([ds, idepth_ds])

    # add 'sst' variable in degrees Kelvin
    ds["sst"] = ds[ocean_names.surface_temperature].copy() + 273.15
    ds["sst"].attrs["long_name"] = "Sea surface temperature"
    ds["sst"].attrs["units"] = "K"

    chunks = config.chunking.get_chunks(ocean_names)
    ds = ds.chunk(chunks)
    ds.attrs["history"] = (
        "Dataset computed by full-model/scripts/data_process"
        "/compute_ocean_dataset.py"
        f" script, using following input sources: {urls}."
    )

    drop_dims = [x for x in list(ds.dims) if x not in ocean_names.full_field_dims]
    ds = ds.drop_dims(drop_dims)
    ds = config.rename(ds)

    return ds


@click.command()
@click.option("--config", help="Path to dataset configuration YAML file.")
@click.option("--run-directory", help="Path to reference run directory.")
@click.option("--output-store", help="Path to output zarr store.")
@click.option(
    "--debug",
    is_flag=True,
    help="Print metadata and return QC plots instead of writing output.",
)
@click.option("--subsample", is_flag=True, help="Subsample the data before writing.")
def main(
    config,
    run_directory,
    output_store,
    debug,
    subsample,
):
    print(f"--run-directory is {run_directory}")
    print(f"--output-store is {output_store}")

    config = OceanDatasetConfig.from_file(config)

    if not debug and config.dask is not None:
        print("Using dask cluster...")
        client = config.dask.get_client()
        print(client)
        print(client.dashboard_link)

    ds = compute_lazy_dataset(
        config=config.dataset_computation,
        run_directory=run_directory,
        fs=config.filesystem,
    )

    if subsample:
        ds = ds.isel(time=slice(None, 73))

    print(f"Output dataset size is {ds.nbytes / 1e9} GB")

    if debug:
        with xr.set_options(display_max_rows=500):
            print(ds)
    else:
        n_partitions = config.n_split
        ds.partition.initialize_store(output_store)
        for i in range(n_partitions):
            print(f"Writing segment {i + 1} / {n_partitions}")
            with ProgressBar():
                ds.partition.write(
                    output_store,
                    n_partitions,
                    ["time"],
                    i,
                    collect_variable_writes=True,
                )


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)
        pdb.post_mortem()  # Start the debugger
        raise  # Re-raise the exception to preserve the traceback



from .atmosphere_data import AtmosphereData
from .device import get_device, using_gpu
from .gridded_ops import GriddedOperations
from .metrics import (
    root_mean_squared_error,
    spherical_area_weights,
    weighted_mean,
    weighted_mean_bias,
    weighted_nanmean,
    weighted_sum,
)
from .normalizer import StandardNormalizer, get_normalizer
from .packer import Packer
from .registry import Registry

__all__ = [
    "spherical_area_weights",
    "weighted_mean",
    "weighted_mean_bias",
    "weighted_nanmean",
    "root_mean_squared_error",
    "get_device",
    "using_gpu",
    "StandardNormalizer",
    "get_normalizer",
    "Packer",
    "AtmosphereData",
    "GriddedOperations",
    "Registry",
]



import dataclasses
import logging
from collections.abc import Sequence
from math import ceil

import cftime
import numpy as np
import torch
import xarray as xr

from fme.ace.data_loading.batch_data import BatchData
from fme.ace.data_loading.perturbation import SSTPerturbation
from fme.ace.requirements import DataRequirements
from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset.config import MergeNoConcatDatasetConfig, XarrayDataConfig
from fme.core.dataset.getters import get_per_dataset_names
from fme.core.dataset.merged import MergedXarrayDataset
from fme.core.dataset.properties import DatasetProperties
from fme.core.dataset.xarray import XarrayDataset
from fme.core.distributed import Distributed
from fme.core.typing_ import Slice


@dataclasses.dataclass
class TimestampList:
    """
    Configuration for a list of timestamps.

    Parameters:
        times: List of timestamps.
        timestamp_format: Format of the timestamps.
    """

    times: Sequence[str]
    timestamp_format: str = "%Y-%m-%dT%H:%M:%S"

    def as_indices(self, time_index: xr.CFTimeIndex) -> np.ndarray:
        datetimes = [
            cftime.datetime.strptime(
                t, self.timestamp_format, calendar=time_index.calendar
            )
            for t in self.times
        ]
        (indices,) = time_index.isin(datetimes).nonzero()
        if len(indices) != len(self.times):
            missing_times = set(datetimes) - set(time_index[indices])
            raise ValueError(
                f"Inference initial condition timestamps {missing_times} "
                "were not found in the dataset."
            )
        return indices

    @property
    def n_initial_conditions(self) -> int:
        return len(self.times)


@dataclasses.dataclass
class InferenceInitialConditionIndices:
    """
    Configuration of the indices for initial conditions during inference.

    Parameters:
        n_initial_conditions: Number of initial conditions to use.
        first: Index of the first initial condition.
        interval: Interval between initial conditions.
    """

    n_initial_conditions: int
    first: int = 0
    interval: int = 1

    def __post_init__(self):
        if self.interval < 0:
            raise ValueError("interval must be positive")

    def as_indices(self) -> np.ndarray:
        stop = self.n_initial_conditions * self.interval + self.first
        return np.arange(self.first, stop, self.interval)


@dataclasses.dataclass
class ExplicitIndices:
    """
    Configure indices providing them explicitly.

    Parameters:
        list: List of integer indices.
    """

    list: Sequence[int]

    def as_indices(self) -> np.ndarray:
        return np.array(self.list)

    @property
    def n_initial_conditions(self) -> int:
        return len(self.list)


@dataclasses.dataclass
class InferenceDataLoaderConfig:
    """
    Configuration for inference data.

    This is like the `DataLoaderConfig` class, but with some additional
    constraints. During inference, we have only one batch, so the number of
    samples directly determines the size of that batch.

    Parameters:
        dataset: Configuration to define the dataset.
        start_indices: Configuration of the indices for initial conditions
            during inference. This can be a list of timestamps, a list of
            integer indices, or a slice configuration of the integer indices.
            Values following the initial condition will still come from
            the full dataset.
        num_data_workers: Number of parallel workers to use for data loading.
        perturbations: Configuration for SST perturbations.
        persistence_names: Names of variables for which all returned values
            will be the same as the initial condition. When evaluating initial
            condition predictability, set this to forcing variables that should
            not be updated during inference (e.g. surface temperature).
    """

    dataset: XarrayDataConfig | MergeNoConcatDatasetConfig
    start_indices: InferenceInitialConditionIndices | ExplicitIndices | TimestampList
    num_data_workers: int = 0
    perturbations: SSTPerturbation | None = None
    persistence_names: Sequence[str] | None = None

    def __post_init__(self):
        if isinstance(self.dataset, XarrayDataConfig):
            if self.dataset.subset != Slice(None, None, None):
                raise ValueError("Inference data may not be subset.")
        elif isinstance(self.dataset, MergeNoConcatDatasetConfig):
            for data in self.dataset.merge:
                if data.subset != Slice(None, None, None):
                    raise ValueError(f"Inference data may not be subset.")
        self._zarr_engine_used = self.dataset.zarr_engine_used

    @property
    def n_initial_conditions(self) -> int:
        return self.start_indices.n_initial_conditions

    @property
    def zarr_engine_used(self) -> bool:
        """
        Whether any of the configured datasets are using the Zarr engine.
        """
        return self._zarr_engine_used


@dataclasses.dataclass
class ForcingDataLoaderConfig:
    """
    Configuration for the forcing data.

    Parameters:
        dataset: Configuration to define the dataset.
        num_data_workers: Number of parallel workers to use for data loading.
        perturbations: Configuration for SST perturbations
            used in forcing data.
        persistence_names: Names of variables for which all returned values
            will be the same as the initial condition. When evaluating initial
            condition predictability, set this to forcing variables that should
            not be updated during inference (e.g. surface temperature).
    """

    dataset: XarrayDataConfig | MergeNoConcatDatasetConfig
    num_data_workers: int = 0
    perturbations: SSTPerturbation | None = None
    persistence_names: Sequence[str] | None = None

    def __post_init__(self):
        if isinstance(self.dataset, XarrayDataConfig):
            if self.dataset.subset != Slice(None, None, None):
                raise ValueError("Inference data may not be subset.")
        elif isinstance(self.dataset, MergeNoConcatDatasetConfig):
            for data in self.dataset.merge:
                if data.subset != Slice(None, None, None):
                    raise ValueError(f"Inference data may not be subset.")

    def build_inference_config(self, start_indices: ExplicitIndices):
        return InferenceDataLoaderConfig(
            dataset=self.dataset,
            num_data_workers=self.num_data_workers,
            start_indices=start_indices,
            perturbations=self.perturbations,
            persistence_names=self.persistence_names,
        )


class InferenceDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        config: InferenceDataLoaderConfig,
        total_forward_steps: int,
        requirements: DataRequirements,
        surface_temperature_name: str | None = None,
        ocean_fraction_name: str | None = None,
    ):
        if isinstance(config.dataset, XarrayDataConfig):
            dataset = XarrayDataset(
                config.dataset, requirements.names, requirements.n_timesteps
            )
            self._dataset = dataset
        elif isinstance(config.dataset, MergeNoConcatDatasetConfig):
            self._dataset = self._resolve_merged_datasets(config.dataset, requirements)
        self._properties = self._dataset.properties
        self._forward_steps_in_memory = requirements.n_timesteps - 1
        self._total_forward_steps = total_forward_steps
        self._perturbations = config.perturbations
        self._surface_temperature_name = surface_temperature_name
        self._ocean_fraction_name = ocean_fraction_name
        self._n_initial_conditions = config.n_initial_conditions
        if isinstance(config.start_indices, TimestampList):
            self._start_indices = config.start_indices.as_indices(
                self._dataset.all_times
            )
        else:
            self._start_indices = config.start_indices.as_indices()
        self._validate_n_forward_steps()
        if isinstance(self._properties.horizontal_coordinates, LatLonCoordinates):
            self._lats, self._lons = self._properties.horizontal_coordinates.meshgrid
        else:
            if self._perturbations is not None:
                raise ValueError(
                    "Currently, SST perturbations are only supported \
                    for lat/lon coordinates."
                )
        if self._perturbations is not None and (
            self._surface_temperature_name is None or self._ocean_fraction_name is None
        ):
            raise ValueError(
                "No ocean configuration found, \
                SST perturbations require an ocean configuration."
            )

        self._persistence_data: BatchData | None = None
        if config.persistence_names is not None:
            first_sample = self._get_batch_data(0)
            self._persistence_data = first_sample.subset_names(
                config.persistence_names
            ).select_time_slice(slice(0, 1))

    def _get_batch_data(self, index) -> BatchData:
        dist = Distributed.get_instance()
        i_start = index * self._forward_steps_in_memory
        sample_tuples = []
        for i_member in range(self._n_initial_conditions):
            # check if sample is one this local rank should process
            if i_member % dist.world_size != dist.rank:
                continue
            i_window_start = i_start + self._start_indices[i_member]
            i_window_end = i_window_start + self._forward_steps_in_memory + 1
            if i_window_end > (
                self._total_forward_steps + self._start_indices[i_member]
            ):
                i_window_end = (
                    self._total_forward_steps + self._start_indices[i_member] + 1
                )
            window_time_slice = slice(i_window_start, i_window_end)
            tensors, time = self._dataset.get_sample_by_time_slice(window_time_slice)
            if self._perturbations is not None:
                if (
                    self._surface_temperature_name is None
                    or self._ocean_fraction_name is None
                ):
                    raise ValueError(
                        "Surface temperature and ocean fraction names must be provided \
                        to apply SST perturbations."
                    )
                logging.debug("Applying SST perturbations to forcing data")
                for perturbation in self._perturbations.perturbations:
                    perturbation.apply_perturbation(
                        tensors[self._surface_temperature_name],
                        self._lats,
                        self._lons,
                        tensors[self._ocean_fraction_name],
                    )
            sample_tuples.append((tensors, time))
        return BatchData.from_sample_tuples(
            sample_tuples,
            horizontal_dims=list(self.properties.horizontal_coordinates.dims),
        )

    def __getitem__(self, index) -> BatchData:
        dist = Distributed.get_instance()
        result = self._get_batch_data(index)
        if self._persistence_data is not None:
            updated_data = {}
            for key, value in self._persistence_data.data.items():
                updated_data[key] = value.expand_as(result.data[key])
            result.data = {**result.data, **updated_data}
        assert result.time.shape[0] == self._n_initial_conditions // dist.world_size
        return result

    def __len__(self) -> int:
        # The ceil is necessary so if the last batch is smaller
        # than the rest the ratio will be rounded up and the last batch
        # will be included in the loading
        return int(ceil(self._total_forward_steps / self._forward_steps_in_memory))

    @property
    def properties(self) -> DatasetProperties:
        return self._properties

    @property
    def n_forward_steps(self) -> int:
        return self._total_forward_steps

    def _validate_n_forward_steps(self):
        max_steps = self._dataset.total_timesteps - max(self._start_indices) - 1
        if self._total_forward_steps > max_steps:
            raise ValueError(
                f"The number of forward inference steps ({self._total_forward_steps}) "
                "must be less than or equal to the number of possible steps "
                f"({max_steps}) in dataset after the last initial condition's "
                "start index."
            )

    def _resolve_merged_datasets(
        self,
        merged_datasets_config: MergeNoConcatDatasetConfig,
        requirements: DataRequirements,
    ):
        per_dataset_names = get_per_dataset_names(
            merged_datasets_config, requirements.names
        )
        merged_xarray_datasets = []
        config_counter = 0
        for config in merged_datasets_config.merge:
            current_dataset = XarrayDataset(
                config,
                per_dataset_names[config_counter],
                requirements.n_timesteps,
            )
            merged_xarray_datasets.append(current_dataset)
            config_counter += 1
        merged_datasets = MergedXarrayDataset(datasets=merged_xarray_datasets)
        return merged_datasets



import logging
import os
import time
from collections.abc import Mapping
from typing import Any

import numpy as np
import wandb

from fme.core.distributed import Distributed

WANDB_RUN_ID_FILE = "wandb_run_id"


class DirectInitializationError(RuntimeError):
    pass


class Histogram(wandb.Histogram):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Histogram.__doc__ = wandb.Histogram.__doc__
Histogram.__init__.__doc__ = wandb.Histogram.__init__.__doc__


class Table(wandb.Table):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Table.__doc__ = wandb.Table.__doc__
Table.__init__.__doc__ = wandb.Table.__init__.__doc__


class Video(wandb.Video):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Video.__doc__ = wandb.Video.__doc__
Video.__init__.__doc__ = wandb.Video.__init__.__doc__


class Image(wandb.Image):
    def __init__(
        self,
        *args,
        direct_access=True,
        **kwargs,
    ):
        if direct_access:
            raise DirectInitializationError(
                "must initialize from `wandb = WandB.get_instance()`, "
                "not directly from `import fme.core.wandb`"
            )
        super().__init__(*args, **kwargs)


Image.__doc__ = wandb.Image.__doc__
Image.__init__.__doc__ = wandb.Image.__init__.__doc__


class WandB:
    """
    A singleton class to interface with Weights and Biases (WandB).
    """

    @classmethod
    def get_instance(cls) -> "WandB":
        """
        Get the singleton instance of the WandB class.
        """
        global singleton
        if singleton is None:
            singleton = cls()
        return singleton

    def __init__(self):
        self._enabled = False
        self._configured = False
        self._id = None

    def configure(self, log_to_wandb: bool):
        dist = Distributed.get_instance()
        self._enabled = log_to_wandb and dist.is_root()
        self._configured = True

    def init(
        self,
        resumable: bool = False,
        experiment_dir: str | None = None,
        **kwargs,
    ):
        """
        Initialize wandb, potentially with resumption logic.

        Args:
            resumable: If True, attempt to resume the run in the experiment directory,
                or start a new run there if no run is found.
            experiment_dir: The directory where the experiment is being run. Required if
                `resumable` is True.
            **kwargs: Passed to wandb.init.
        """
        if not self._configured:
            raise RuntimeError(
                "must call WandB.configure before WandB init can be called"
            )
        if self._enabled:
            if resumable:
                if experiment_dir is None:
                    raise ValueError(
                        "must provide `experiment_dir` when `resumable` is True"
                    )
                else:
                    id_ = init_wandb_with_resumption(
                        experiment_dir, direct_access=False, **kwargs
                    )
            else:
                wandb.init(**kwargs)
                if wandb.run is None:
                    raise RuntimeError("wandb.init did not return a run")
                else:
                    id_ = wandb.run.id
                logging.info(f"New non-resuming wandb run with id: {id_}.")
            self._id = id_

    def watch(self, modules):
        if self._enabled:
            wandb.watch(modules)

    def log(self, data: Mapping[str, Any], step=None, sleep=None):
        if self._enabled:
            wandb.log(dict(data), step=step)
            if sleep is not None:
                time.sleep(sleep)

    def Image(self, data_or_path, *args, **kwargs) -> Image:
        if isinstance(data_or_path, np.ndarray):
            data_or_path = scale_image(data_or_path)

        return Image(data_or_path, *args, direct_access=False, **kwargs)

    def Video(self, *args, **kwargs) -> Video:
        return Video(*args, direct_access=False, **kwargs)

    def Table(self, *args, **kwargs) -> Table:
        return Table(*args, direct_access=False, **kwargs)

    def Histogram(self, *args, **kwargs) -> Histogram:
        return Histogram(*args, direct_access=False, **kwargs)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def configured(self) -> bool:
        return self._configured

    def get_id(self) -> str | None:
        return self._id


singleton: WandB | None = None


def scale_image(
    image_data: np.ndarray,
) -> np.ndarray:
    """
    Given an array of scalar data, rescale the data to the range [0, 255].
    """
    data_min = np.nanmin(image_data)
    data_max = np.nanmax(image_data)
    # video data is brightness values on a 0-255 scale
    image_data = 255 * (image_data - data_min) / (data_max - data_min)
    image_data = np.minimum(image_data, 255)
    image_data = np.maximum(image_data, 0)
    image_data[np.isnan(image_data)] = 0
    return image_data


def init_wandb_with_resumption(
    experiment_dir: str,
    direct_access=True,
    wandb_run_id_file: str = WANDB_RUN_ID_FILE,
    wandb_init=None,
    wandb_id=None,
    **kwargs: Any,
) -> str:
    """
    Initialize wandb with resumption logic. If wandb has previously
    been initialized in the experiment directory, resume the run. Otherwise,
    start a new run.

    The reason we implement our own resumption logic is that wandb uses the same
    location to save temporary media files and the information necessary for
    resumption. We want to save these things in different places.

    Args:
        experiment_dir: The directory where the experiment is being run.
        direct_access: If True, raise an error if this function is called directly.
        wandb_run_id_file: The file where the wandb run id is stored.
        wandb_init: The wandb.init function to use (for testing).
        wandb_id: A function returning the wandb run_id (for testing).
        **kwargs: Arguments to pass to `wandb.init`.

    Returns:
        The wandb run id.
    """
    if direct_access:
        raise DirectInitializationError(
            "Must access this function by calling `wandb.init` after "
            "`wandb = WandB.get_instance()`. It should not be called from anywhere "
            "else."
        )

    if wandb_init is None:
        wandb_init = wandb.init

    if wandb_id is None:

        def wandb_id():
            if wandb.run is None:
                raise RuntimeError("wandb does not have an active run")
            return wandb.run.id

    if not os.path.exists(os.path.join(experiment_dir, wandb_run_id_file)):
        # new run
        kwargs.update({"resume": "never"})
        wandb_init(**kwargs)
        logging.info(f"New resumable wandb run with id: {wandb_id()}.")
        with open(os.path.join(experiment_dir, wandb_run_id_file), "w") as f:
            f.write(wandb_id())
    else:
        # resuming
        with open(os.path.join(experiment_dir, wandb_run_id_file)) as f:
            wandb_run_id = f.read().strip()
        kwargs.update({"resume": "must", "id": wandb_run_id})
        wandb_init(**kwargs)
        logging.info(f"Resuming wandb run with id: {wandb_id()}")
    return wandb_id()



from fme.core.typing_ import TensorMapping


class NullAggregator:
    """
    An aggregator that does nothing. Null object pattern.
    """

    def __init__(self):
        pass

    def record_batch(
        self,
        loss: float,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ):
        pass

    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        return {}



import contextlib

import pytest
import torch
from torch import nn

from fme.core.wildcard import apply_by_wildcard, wildcard_match


@pytest.mark.parametrize(
    "pattern, name, expected",
    [
        ("*", "abc", True),
        ("*", "abc.def", True),
        ("abc*", "abc", True),
        ("abc*", "abc.def", True),
        ("abc*", "def", False),
        ("abc.*", "abc.def", True),
        ("*.abc.*", "abc.def", False),
        ("*.def.*", "abc.def", False),
        ("*.def.*", "abc.def.ghi", True),
        ("abc.*", "abc.def.ghi", True),
        ("abc.*.ghi", "abc.def.ghi", True),
        ("abc.*.ghi", "abc.def", False),
        ("abc.*.ghi", "abc.def.ghi.jkl", False),
        ("*.abc.ghi", "def.abc.ghi", True),
        ("*.abc.ghi", "abc.ghi", False),
        ("*.abc.ghi", "def.abc.ghi.jkl", False),
    ],
)
def test_wildcard_match(pattern, name, expected):
    assert wildcard_match(pattern=pattern, name=name) == expected


class NestedModule2(nn.Module):
    def __init__(self):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(3, 3))


class NestedModule1(nn.Module):
    def __init__(self):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(3, 3))
        self.nested = NestedModule2()


@pytest.mark.parametrize(
    "include, exclude, expected_applied, expected_error",
    [
        pytest.param(["*"], [], ["weight", "nested.weight"], None, id="include all"),
        pytest.param([], ["*"], [], None, id="exclude all"),
        pytest.param(["weight"], ["nested.*"], ["weight"], None, id="weight included"),
        pytest.param(["*"], ["nested.*"], [], ValueError, id="nested param in both"),
        pytest.param(["*"], ["weight"], [], ValueError, id="* include with an exclude"),
        pytest.param([], ["weight"], [], ValueError, id="missing weight using exclude"),
        pytest.param(["weight"], [], [], ValueError, id="missing weight using include"),
        pytest.param(
            ["*.weight"], [], [], ValueError, id="mising weight using wildcard include"
        ),
    ],
)
def test_apply_by_wildcard(
    include: list[str],
    exclude: list[str],
    expected_applied: list[str],
    expected_error: type[Exception] | None,
):
    model = NestedModule1()

    def func(module: nn.Module, name: str):
        module.get_parameter(name).requires_grad = False

    if expected_error is not None:
        context = pytest.raises(expected_error)
    else:
        context = contextlib.nullcontext()

    with context:
        apply_by_wildcard(
            model,
            func,
            include,
            exclude,
        )

    if expected_error is None:
        for name, param in model.named_parameters():
            assert param.requires_grad == (name not in expected_applied)



import abc
import re
from collections.abc import Callable
from typing import Any, TypeVar

import torch

from fme.core.masking import NullMasking, StaticMasking
from fme.core.typing_ import TensorDict, TensorMapping

LEVEL_PATTERN = re.compile(r"_(\d+)$")


class MaskProviderABC(abc.ABC):
    SelfType = TypeVar("SelfType", bound="MaskProviderABC")

    @abc.abstractmethod
    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None: ...

    @abc.abstractmethod
    def to(self: SelfType, device: str) -> SelfType: ...

    @abc.abstractmethod
    def update(self: SelfType, other: SelfType) -> None: ...

    @abc.abstractmethod
    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]: ...

    @abc.abstractmethod
    def to_state(self) -> dict[str, Any]: ...


class _NullMaskProvider(MaskProviderABC):
    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        return None

    def to(self, device: str) -> "_NullMaskProvider":
        return self

    def update(self, other: MaskProviderABC) -> None:
        if not isinstance(other, _NullMaskProvider):
            raise ValueError(
                f"Attempted to update NullMaskProvider with non-null {other}"
            )

    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]:
        return NullMasking()

    def to_state(self) -> dict[str, Any]:
        return {"masks": {}}


NullMaskProvider = _NullMaskProvider()


class MaskProvider(MaskProviderABC):
    """
    Stores and returns 2D mask tensors.

    Masks are special time-invariant data variables with names that start with
    the string "mask_". There are three types of masks that MaskProvider knows
    how to use, in order from highest to lowest priority:

    1. Variable-specific masks, e.g. "mask_sst" and "mask_thetao_1".
    2. Level-specific masks, e.g. "mask_0" and "mask_1".
    3. A catch-all 2D mask named "mask_2d".

    For example, when matching the variable "theta_1", MaskProvider will first
    check if it has a mask called "mask_theta_1". If not, then it will check for
    a mask called "mask_1", returning None if this is also not found.

    Another example: when matching the variable "sst", MaskProvider checks for
    "mask_sst" and then "mask_2d", returning None if neither are found.

    Args:
        masks: A dictionary where the keys are required to start with "mask_"
            and values are 2D mask tensors.

    """

    def __init__(self, masks: TensorDict | None = None):
        self._masks = {} if masks is None else masks
        for key in self._masks:
            if not key.startswith("mask_"):
                raise ValueError(
                    "The 'mask' TensorDict passed to MaskProvider init has "
                    f"non-mask tensors, including {key}. Expected all keys "
                    "to start with the string 'mask_'."
                )

    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]:
        """
        Returns a StaticMasking object that fills in NaNs outside of mask
        valid points, i.e. where the mask value is 0.

        """
        return StaticMasking(
            mask_value=0,
            fill_value=float("nan"),
            mask=self,
        )

    @property
    def masks(self) -> TensorMapping:
        return self._masks

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        # variable specific
        mask_name = f"mask_{name}"
        if mask_name in self.masks:
            return self.masks[mask_name]
        # level specific for 3D vars
        match = LEVEL_PATTERN.search(name)
        if match:
            level = int(match.group(1))
            mask_name = f"mask_{level}"
            return self.masks.get(mask_name, None)
        # 2D mask or None
        return self.masks.get("mask_2d", None)

    def to(self, device: str) -> "MaskProvider":
        return MaskProvider(
            {name: tensor.to(device) for name, tensor in self.masks.items()}
        )

    def update(self, other: "MaskProvider") -> None:
        """Update the MaskProvider's masks with masks from another MaskProvider.

        Raises a ValueError if there are overlapping mask names between the two
        MaskProviders.
        """
        self_keys = set(self.masks.keys())
        other_keys = set(other.masks.keys())
        intersection = self_keys.intersection(other_keys)
        if intersection:
            raise ValueError(
                "Cannot update MaskProvider with overlapping mask names: "
                f"{', '.join(sorted(list(intersection)))}"
            )
        self._masks.update(other.masks)

    def __eq__(self, other) -> bool:
        if not isinstance(other, MaskProvider):
            return False
        if not self.masks.keys() == other.masks.keys():
            return False
        for name, mask in self.masks.items():
            try:
                torch.testing.assert_close(mask, other.masks[name])
            except AssertionError:
                return False
        return True

    def __repr__(self) -> str:
        return (
            "MaskProvider(\n    masks=[\n        "
            + ",\n        ".join(sorted(list(self.masks.keys())))
            + ",\n    ]\n)"
        )

    def to_state(self) -> dict[str, Any]:
        return {"masks": self.masks}

    @classmethod
    def from_state(cls, state: dict[str, Any]) -> "MaskProvider":
        return cls(masks=state["masks"])



from collections.abc import Mapping
from typing import Any


def to_flat_dict(d: Mapping[str, Any]) -> dict[str, Any]:
    """
    Converts any nested dictionaries to a flat version with
    the nested keys joined with a '.', e.g., {a: {b: 1}} ->
    {a.b: 1}.
    """
    new_flat = {}
    for k, v in d.items():
        if isinstance(v, dict):
            sub_d = to_flat_dict(v)
            for sk, sv in sub_d.items():
                new_flat[".".join([k, sk])] = sv
        else:
            new_flat[k] = v

    return new_flat


def to_nested_dict(d: Mapping[str, Any]) -> dict[str, Any]:
    """
    Converts a flat dictionary with '.' joined keys back into
    a nested dictionary, e.g., {a.b: 1} -> {a: {b: 1}}.
    """
    new_config: dict[str, Any] = {}

    for k, v in d.items():
        if "." in k:
            sub_keys = k.split(".")
            sub_d = new_config
            for sk in sub_keys[:-1]:
                sub_d = sub_d.setdefault(sk, {})
            sub_d[sub_keys[-1]] = v
        else:
            new_config[k] = v

    return new_config


def add_names(
    left: Mapping[str, Any], right: Mapping[str, Any], names: list[str]
) -> dict[str, Any]:
    """Add the 'names' from left dict to the right dict and return result.

    Args:
        left: These values will be added to the right dict.
        right: The dict to add the values to.
        names: The names of the keys to add.

    Returns:
        A new dict with the same keys as 'right', but with the values from 'left' added
        for specified names.
    """
    return {k: left[k] + right[k] if k in names else right[k] for k in right}



import pytest

from fme.ace.registry.m2lines import SamudraBuilder


def test_samudra_builder():
    builder = SamudraBuilder()
    # assuming 5 input (3 prognostic + 2 forcing) and 3 output vars (prognostic)
    model = builder.build(5, 3, (16, 32))
    assert model.layers[0].convblock[0].in_channels == 5
    assert model.layers[-1].out_channels == 3

    with pytest.raises(ValueError, match="norm_kwargs should not have num_features"):
        _ = SamudraBuilder(norm_kwargs={"num_features": 10})

    with pytest.raises(
        ValueError, match="norm_kwargs should not have normalized_shape"
    ):
        _ = SamudraBuilder(norm_kwargs={"normalized_shape": (3, 3)})



import os

import pytest
import torch

from fme.core.device import get_device
from fme.core.typing_ import TensorDict


def validate_tensor(x: torch.Tensor, filename: str):
    if not os.path.exists(filename):
        torch.save(x.cpu(), filename)
        pytest.fail(f"Regression file {filename} did not exist, so it was created")
    else:
        y = torch.load(filename, map_location=get_device()).to(x.device)
        torch.testing.assert_close(x, y)


def validate_tensor_dict(x: TensorDict, filename: str):
    if not os.path.exists(filename):
        x_cpu = {k: v.cpu() for k, v in x.items()}
        torch.save(x_cpu, filename)
        pytest.fail(f"Regression file {filename} did not exist, so it was created")
    else:
        y = torch.load(filename, map_location=get_device())
        for k, v in x.items():
            torch.testing.assert_close(v, y[k].to(v.device))



"""
Create a new netcdf file with the specified variables repeated n times in time.

Example usage:

```
python compute_repeating_forcing.py \
    --input_dir /net/nfs/climate/data/2023-08-11-vertically-resolved-1deg-fme-ensemble-dataset/validation/ic_0011 \
    --output_dir /tmp/repeating_climate/ \
    --n_times 2 \
    -f DSWRFtoa \
    -f surface_temperature \
    -f ocean_fraction \
    -f land_fraction \
    -f sea_ice_fraction
```
"""  # noqa: E501

from glob import glob
from pathlib import Path
from typing import Dict

import click
import numpy as np
import xarray as xr


def get_time_invariant_vars(ds: xr.Dataset) -> Dict[str, xr.DataArray]:
    arrays = {}
    for var in ds.data_vars:
        if "time" not in ds[var].dims:
            arrays[str(var)] = ds[var]
    return arrays


@click.command()
@click.option(
    "--n_times",
    type=int,
    help=(
        "Total number of repeats, e.g. 2-times means that the original forcing "
        "is there plus and additional repeat."
    ),
    required=True,
)
@click.option(
    "--input_dir",
    type=Path,
    help=("Directory containing netcdf files, e.g. from a single ensemble member"),
    required=True,
)
@click.option(
    "--output_dir",
    type=Path,
    help="Where to store the output netcdf files.",
    required=True,
)
@click.option(
    "-f",
    "--repeat_variables",
    help="Variables to repeat (usually forcing and prescribed variables)",
    default=["surface_temperature"],
    multiple=True,
)
@click.option("--nc-format", default="NETCDF3_64BIT", help="netCDF file format")
def main(n_times: int, input_dir: Path, output_dir: Path, repeat_variables, nc_format):
    click.echo("Loading data...", nl=False)
    files = sorted(glob(str(input_dir / "*.nc")))
    arrays = get_time_invariant_vars(xr.open_dataset(files[0], decode_timedelta=False))
    ds = xr.open_mfdataset(files)
    click.echo("done")

    click.echo("Detected time-invariant variables:")
    for name in arrays:
        click.echo(f'"{name}": {arrays[name].shape}')

    # initial condition
    ic_vars = []
    for var in ds.data_vars:
        if var not in repeat_variables and var not in arrays:
            ic_vars.append(str(var))
            arrays[str(var)] = (
                ds[var]
                .isel(time=0)
                .expand_dims("initial_condition", axis=0)
                .copy(deep=True)
            )

    time_axis = 0
    for var in repeat_variables:
        click.echo(f'Repeating "{var}"')
        arrays[var] = xr.DataArray(  # type: ignore
            np.repeat(ds[var].data, n_times, axis=time_axis), dims=ds[var].dims
        )

    dt: int = (ds.time[1] - ds.time[0]).values.astype("timedelta64[h]").astype(int)
    time_coord = xr.date_range(
        ds.time.item(0),
        periods=len(ds.time) * n_times,
        freq=f"{dt}h",
        calendar=ds.time.dt.calendar,
        use_cftime=True,
    )

    coords = {**ds.coords, "time": time_coord, "initial_condition": [0]}
    monthly_ds = xr.Dataset(arrays, coords=coords).resample(time="MS")
    bytes = 0.0
    history = (
        f"Created by running "
        f"`full_model/project/data_process/compute_repeating_forcing.py --input_dir "
        f"{input_dir} -f {' -f '.join(repeat_variables)}`"
    )
    for i, (label, data) in enumerate(monthly_ds):
        if i != 0:
            data = data.drop_vars(ic_vars)
        bytes += data.nbytes
        click.echo(
            f"Processing month {i + 1}/{len(monthly_ds)} ({(bytes / 1e9):.2f} GB)\r",
            nl=False,
        )
        # use these options to enable opening data with netCDF4.MFDataset
        outpath = (output_dir / label.strftime("%Y%m%d%H")).with_suffix(".nc")
        data.attrs["history"] = history
        data.to_netcdf(outpath, unlimited_dims=["time"], format=nc_format)

    click.echo("")


if __name__ == "__main__":
    main()  # type: ignore



from fme.core.dataset.data_typing import VariableMetadata

ERA5_VERSION_1 = {
    "DLWRFsfc": {
        "long_name": "Downward LW radiative flux at surface",
        "units": "W/m**2",
    },
    "DPT2m": {"long_name": "2m dewpoint temperature", "units": "K"},
    "DSWRFsfc": {
        "long_name": "Downward SW radiative flux at surface",
        "units": "W/m**2",
    },
    "DSWRFtoa": {"long_name": "Downward SW radiative flux at TOA", "units": "W/m**2"},
    "HGTsfc": {"long_name": "Topography height", "units": "m"},
    "LHTFLsfc": {"long_name": "Latent heat flux", "units": "W/m**2"},
    "PRATEsfc": {"long_name": "Surface precipitation rate", "units": "kg/m**2/s"},
    "PRESsfc": {"long_name": "Surface pressure", "units": "Pa"},
    "Q200": {"long_name": "Specific humidity at 200 hPa", "units": "kg kg**-1"},
    "Q2m": {"long_name": "2m specific humidity", "units": "kg/kg"},
    "Q500": {"long_name": "Specific humidity at 500 hPa", "units": "kg kg**-1"},
    "Q850": {"long_name": "Specific humidity at 850 hPa", "units": "kg kg**-1"},
    "SHTFLsfc": {"long_name": "Sensible heat flux", "units": "W/m**2"},
    "TMP200": {"long_name": "Temperature at 200 hPa", "units": "K"},
    "TMP2m": {"long_name": "2m air temperature", "units": "K"},
    "TMP500": {"long_name": "Temperature at 500 hPa", "units": "K"},
    "TMP850": {"long_name": "Temperature at 850 hPa", "units": "K"},
    "UGRD10m": {"long_name": "10m U component of wind", "units": "m/s"},
    "UGRD200": {"long_name": "U component of wind at 200 hPa", "units": "m s**-1"},
    "UGRD500": {"long_name": "U component of wind at 500 hPa", "units": "m s**-1"},
    "UGRD850": {"long_name": "U component of wind at 850 hPa", "units": "m s**-1"},
    "ULWRFsfc": {"long_name": "Upward LW radiative flux at surface", "units": "W/m**2"},
    "ULWRFtoa": {"long_name": "Upward LW radiative flux at TOA", "units": "W/m**2"},
    "USWRFsfc": {"long_name": "Upward SW radiative flux at surface", "units": "W/m**2"},
    "USWRFtoa": {"long_name": "Upward SW radiative flux at TOA", "units": "W/m**2"},
    "VGRD10m": {"long_name": "10m V component of wind", "units": "m/s"},
    "VGRD200": {"long_name": "V component of wind at 200 hPa", "units": "m s**-1"},
    "VGRD500": {"long_name": "V component of wind at 500 hPa", "units": "m s**-1"},
    "VGRD850": {"long_name": "V component of wind at 850 hPa", "units": "m s**-1"},
    "air_temperature_0": {"long_name": "Temperature level-0", "units": "K"},
    "air_temperature_1": {"long_name": "Temperature level-1", "units": "K"},
    "air_temperature_2": {"long_name": "Temperature level-2", "units": "K"},
    "air_temperature_3": {"long_name": "Temperature level-3", "units": "K"},
    "air_temperature_4": {"long_name": "Temperature level-4", "units": "K"},
    "air_temperature_5": {"long_name": "Temperature level-5", "units": "K"},
    "air_temperature_6": {"long_name": "Temperature level-6", "units": "K"},
    "air_temperature_7": {"long_name": "Temperature level-7", "units": "K"},
    "eastward_wind_0": {"long_name": "U component of wind level-0", "units": "m s**-1"},
    "eastward_wind_1": {"long_name": "U component of wind level-1", "units": "m s**-1"},
    "eastward_wind_2": {"long_name": "U component of wind level-2", "units": "m s**-1"},
    "eastward_wind_3": {"long_name": "U component of wind level-3", "units": "m s**-1"},
    "eastward_wind_4": {"long_name": "U component of wind level-4", "units": "m s**-1"},
    "eastward_wind_5": {"long_name": "U component of wind level-5", "units": "m s**-1"},
    "eastward_wind_6": {"long_name": "U component of wind level-6", "units": "m s**-1"},
    "eastward_wind_7": {"long_name": "U component of wind level-7", "units": "m s**-1"},
    "global_mean_co2": {"long_name": "global mean CO2 concentration", "units": "ppm"},
    "h1000": {"long_name": "Geopotential height at 1000 hPa", "units": "m"},
    "h200": {"long_name": "Geopotential height at 200 hPa", "units": "m"},
    "h250": {"long_name": "Geopotential height at 250 hPa", "units": "m"},
    "h300": {"long_name": "Geopotential height at 300 hPa", "units": "m"},
    "h500": {"long_name": "Geopotential height at 500 hPa", "units": "m"},
    "h700": {"long_name": "Geopotential height at 700 hPa", "units": "m"},
    "h850": {"long_name": "Geopotential height at 850 hPa", "units": "m"},
    "land_fraction": {"long_name": "land fraction", "units": "(0-1)"},
    "northward_wind_0": {
        "long_name": "V component of wind level-0",
        "units": "m s**-1",
    },
    "northward_wind_1": {
        "long_name": "V component of wind level-1",
        "units": "m s**-1",
    },
    "northward_wind_2": {
        "long_name": "V component of wind level-2",
        "units": "m s**-1",
    },
    "northward_wind_3": {
        "long_name": "V component of wind level-3",
        "units": "m s**-1",
    },
    "northward_wind_4": {
        "long_name": "V component of wind level-4",
        "units": "m s**-1",
    },
    "northward_wind_5": {
        "long_name": "V component of wind level-5",
        "units": "m s**-1",
    },
    "northward_wind_6": {
        "long_name": "V component of wind level-6",
        "units": "m s**-1",
    },
    "northward_wind_7": {
        "long_name": "V component of wind level-7",
        "units": "m s**-1",
    },
    "ocean_fraction": {"long_name": "ocean fraction", "units": "(0-1)"},
    "sea_ice_fraction": {"long_name": "sea ice fraction", "units": "(0-1)"},
    "soil_moisture_0": {
        "long_name": "Volumetric soil water layer 1",
        "units": "m**3 m**-3",
    },
    "soil_moisture_1": {
        "long_name": "Volumetric soil water layer 2",
        "units": "m**3 m**-3",
    },
    "soil_moisture_2": {
        "long_name": "Volumetric soil water layer 3",
        "units": "m**3 m**-3",
    },
    "soil_moisture_3": {
        "long_name": "Volumetric soil water layer 4",
        "units": "m**3 m**-3",
    },
    "specific_total_water_0": {
        "long_name": "Specific total water level-0",
        "units": "kg kg**-1",
    },
    "specific_total_water_1": {
        "long_name": "Specific total water level-1",
        "units": "kg kg**-1",
    },
    "specific_total_water_2": {
        "long_name": "Specific total water level-2",
        "units": "kg kg**-1",
    },
    "specific_total_water_3": {
        "long_name": "Specific total water level-3",
        "units": "kg kg**-1",
    },
    "specific_total_water_4": {
        "long_name": "Specific total water level-4",
        "units": "kg kg**-1",
    },
    "specific_total_water_5": {
        "long_name": "Specific total water level-5",
        "units": "kg kg**-1",
    },
    "specific_total_water_6": {
        "long_name": "Specific total water level-6",
        "units": "kg kg**-1",
    },
    "specific_total_water_7": {
        "long_name": "Specific total water level-7",
        "units": "kg kg**-1",
    },
    "surface_temperature": {"long_name": "Skin temperature", "units": "K"},
    "tendency_of_total_water_path_due_to_advection": {
        "long_name": "Tendency of total water path due to advection",
        "units": "kg/m**2/s",
    },
    "total_column_water_vapour": {
        "long_name": "Total column vertically-integrated water vapour",
        "units": "kg m**-2",
    },
}

VERSIONS = {
    "era5_v1": ERA5_VERSION_1,
}


def get_default_variable_metadata(
    version: str = "era5_v1",
) -> dict[str, VariableMetadata]:
    asdict = VERSIONS[version]
    return {key: VariableMetadata(**value) for key, value in asdict.items()}
